import numpy as np
import cv2
import matplotlib.pyplot as plt
from scipy.signal import medfilt2d
from scipy.ndimage import gaussian_filter
from scipy.optimize import leastsq as optimize_least_squares
from scipy.optimize import minimize
from scipy.ndimage import sobel


class OptimimzerStitching:
    def __init__(
        self,
        images,
        points,
        sigmas,
        window_size,
        use_gradients=True,
        binarize_grad=True,
    ):
        self.images = images.copy()
        self.points = points.copy()

    def stitch(self):

        # upper left and right
        img_1 = self.images[0].astype(np.float64)
        img_2 = self.images[1].astype(np.float64)
        points_1 = self.points[0]
        points_2 = self.points[1]

        # subtract local mean
        f_1 = medfilt2d(img_1, 31)
        f_2 = medfilt2d(img_2, 31)
        img_1 -= f_1
        img_2 -= f_2

        # get rectification transforms
        target_1 = np.array(
            [(0.0, 0.0), (3.0, 0.0), (0.0, 5.0), (3.0, 5.0)]
        )  # cells have unit length
        target_2 = target_1 + np.array([(3.0, 0.0)])
        H_1, _ = cv2.findHomography(points_1, target_1)
        H_2, _ = cv2.findHomography(points_2, target_2)

        # fix resolution of target
        cell_size = (points_1[1, 0] - points_1[0, 0]) / 3
        S = np.diag([cell_size, cell_size, 1])  # scale matrix

        # Test: Initially warp img_1 and img_2
        # res = self._warp_both(img_1, img_2, H_1, H_2, S, (6,5))
        # plt.imshow(res)
        # plt.show()

        # find ROI in target coordinate system
        lb = np.array([(0.0, 0.0), (0.0, img_2.shape[0])], dtype=np.float64).reshape(
            -1, 1, 2
        )
        rb = np.array(
            [(img_1.shape[1], 0.0), (img_1.shape[1], img_1.shape[0])], dtype=np.float64
        ).reshape(-1, 1, 2)
        img_2_left_border = cv2.perspectiveTransform(lb, H_2).reshape(-1, 2)
        img_1_right_border = cv2.perspectiveTransform(rb, H_1).reshape(-1, 2)
        rect_x1 = max(img_2_left_border[0, 0], img_2_left_border[1, 0])
        rect_y1 = max(img_2_left_border[0, 1], img_1_right_border[0, 1])
        rect_x2 = min(img_1_right_border[0, 0], img_1_right_border[1, 0])
        rect_y2 = min(img_1_right_border[1, 1], img_2_left_border[1, 1])
        roi = np.array(
            [
                (rect_x1, rect_y1),
                (rect_x2, rect_y1),
                (rect_x1, rect_y2),
                (rect_x2, rect_y2),
            ]
        )
        centroid = np.mean(roi, axis=0, keepdims=True)
        print(roi)
        print(centroid)
        roi = (
            roi - centroid
        ) * 0.75 + centroid  # only use inner 75% of ROI so that we can keep that constant during optimization
        roi_w = roi[1, 0] - roi[0, 0]
        roi_h = roi[2, 1] - roi[0, 1]
        roi_area = (roi_w, roi_h)

        # fix H_1 and H_2 so that the top left edge of roi maps to (0, 0)
        fix = np.diag([1.0, 1.0, 1.0])
        fix[:2, 2] = -roi[0]
        H_1_roi = fix.dot(H_1)
        H_2_roi = fix.dot(H_2)

        # Test: plot roi in both images
        self._plot_roi(img_1, img_2, H_1_roi, H_2_roi, S, roi_area)

        # iterate over scales, where s is the amount of downscaling,
        # eg. s=8 means that images are downscaled by a factor of 8
        for s in (8, 4, 2, 1):
            S_scaled = np.diag([1 / s, 1 / s, 1]).dot(S)
            sigma = (
                3 * s / (2 * np.pi)
            )  # standard deviation of filter that cuts at Nyquist frequency according to s
            img_1_f = gaussian_filter(img_1, sigma)
            img_2_f = gaussian_filter(img_2, sigma)

            # initial parameters:
            # parameters give pertubation of 4 corner points of the ROI in the second image
            x_init = np.zeros(8, dtype=np.float64)

            # we keep H_1 fixed, so we can precompute roi_1
            roi_1 = self._warp(img_1_f, H_1_roi, S_scaled, roi_area)

            def opt_fn(x):
                H_corr, _ = cv2.findHomography(roi, roi + x.reshape(4, 2))
                roi_2 = self._warp(img_2_f, H_corr.dot(H_2_roi), S_scaled, roi_area)
                return -np.mean(roi_1 * roi_2)

            # run optimization
            res = minimize(opt_fn, x_init, method="COBYLA", options={"tol": 1e-12})
            print(res)

            # Test: plot final state
            H_corr, _ = cv2.findHomography(roi, roi + res.x.reshape(4, 2))
            if s == 1:
                self._plot_roi(
                    img_1_f, img_2_f, H_1_roi, H_2_roi, S_scaled, roi_area, H_corr
                )

            # update H_2_roi
            H_2_roi = H_corr.dot(H_2_roi)

        # undo shift of H_2
        H_2_old = H_2.copy()
        fix[:2, 2] = roi[0]
        H_2 = fix.dot(H_2_roi)

        result_old = self._warp_both(img_1, img_2, H_1, H_2_old, S, (6, 5))
        result = self._warp_both(img_1, img_2, H_1, H_2, S, (6, 5))
        plt.subplot(1, 2, 1)
        plt.imshow(result_old)
        plt.subplot(1, 2, 2)
        plt.imshow(result)
        plt.show()

        return None

    def _plot_roi(self, img_1, img_2, H_1_roi, H_2_roi, S, area, H_corr=None):
        roi_1 = self._warp(img_1, H_1_roi, S, area)
        roi_2 = self._warp(img_2, H_2_roi, S, area)

        if H_corr is None:
            plt.subplot(1, 4, 1)
            plt.imshow(roi_1)
            plt.subplot(1, 4, 2)
            plt.imshow(roi_2)
            plt.subplot(1, 4, 3)
            plt.imshow(-(roi_1 * roi_2))
            plt.subplot(1, 4, 4)
            plt.imshow(roi_1 - roi_2)
        else:
            roi_res = self._warp(img_2, H_corr.dot(H_2_roi), S, area)
            plt.subplot(2, 4, 1)
            plt.imshow(roi_1)
            plt.subplot(2, 4, 2)
            plt.imshow(roi_2)
            plt.subplot(2, 4, 3)
            plt.imshow(-(roi_1 * roi_2))
            plt.subplot(2, 4, 4)
            plt.imshow(roi_1 - roi_2)

            plt.subplot(2, 4, 5)
            plt.imshow(roi_1)
            plt.subplot(2, 4, 6)
            plt.imshow(roi_res)
            plt.subplot(2, 4, 7)
            plt.imshow(-(roi_1 * roi_res))
            plt.subplot(2, 4, 8)
            plt.imshow(roi_1 - roi_res)

        plt.show()

    def _warp(self, img, H, S, area):
        size = (int(area[0] * S[0, 0]), int(area[1] * S[1, 1]))
        return cv2.warpPerspective(img, S.dot(H), size)

    def _warp_both(self, img_1, img_2, H_1, H_2, S, area):
        size = (int(area[0] * S[0, 0]), int(area[1] * S[1, 1]))
        res1 = self._warp(img_1, H_1, S, area)
        res1[:, size[0] // 2 :] = 0.0
        res2 = self._warp(img_2, H_2, S, area)
        res2[:, : size[0] // 2] = 0.0
        return res1 + res2



from pvinspect.common.util import *
from shapely import *
from test.utilities import *


def test_iou():
    p1 = Polygon([(0, 0), (2, 0), (2, 2), (0, 2)])
    p2 = affinity.translate(p1, -1, -1)
    assert_equal(iou(p1, p2), 1 / 7)


def test_mean_iou():
    p1_1 = Polygon([(0, 0), (2, 0), (2, 2), (0, 2)])
    p1_2 = affinity.translate(p1_1, -1, -1)  # iou: 1/7
    p2_1 = affinity.translate(p1_1, 1.5, 1.5)
    p2_2 = affinity.translate(p2_1, 1, 1)  # iou: 1/7
    iou, _, _ = objdetect_metrics(
        [("Module", p1_1), ("Module", p2_1)], [("Module", p1_2), ("Module", p2_2)]
    )
    assert_equal(iou, (1 / 7 + 1 / 7) / 2)


def test_precision_recall1():
    p1_1 = Polygon([(0, 0), (2, 0), (2, 2), (0, 2)])
    p1_2 = affinity.translate(p1_1, -1, -1)  # iou: 1/7
    p2_1 = affinity.translate(p1_1, 1.5, 1.5)
    p2_2 = affinity.translate(p2_1, 0, 0)  # iou: 1
    p3_2 = affinity.translate(p1_1, 10, 10)  # something far off
    _, prec, recall = objdetect_metrics(
        [("Module", p1_1), ("Module", p2_1)],
        [("Module", p1_2), ("Module", p2_2), ("Module", p3_2)],
        [1 / 8, 0.999],
    )
    assert_equal(prec[0], 2 / 3)
    assert_equal(prec[1], 1 / 3)
    assert_equal(recall[0], 1)
    assert_equal(recall[1], 1 / 2)


def test_precision_recall2():
    p1_1 = Polygon([(0, 0), (2, 0), (2, 2), (0, 2)])
    p1_2 = affinity.translate(p1_1, -1, -1)  # iou: 1/7
    p2_1 = affinity.translate(p1_1, 1.5, 1.5)
    p2_2 = affinity.translate(p2_1, 0, 0)  # iou: 1
    p3_1 = affinity.translate(p1_1, 10, 10)  # something far off
    iou, prec, recall = objdetect_metrics(
        [("Module", p1_1), ("Module", p2_1), ("Module", p3_1)],
        [("Module", p1_2), ("Module", p2_2)],
        [1 / 8, 0.999],
    )
    assert_equal(prec[0], 1)
    assert_equal(prec[1], 1 / 2)
    assert_equal(recall[0], 2 / 3)
    assert_equal(recall[1], 1 / 3)



import onnxruntime
import numpy as np
from scipy import special
from typing import List, Tuple
from pvinspect.data.image import ImageOrSequence, _sequence, Image
from skimage.transform import rescale
from pathlib import Path


class ONNXModel:
    def __init__(
        self,
        file: Path,
        classes: List[str],
        logistic_transform: bool,
        input_shape: Tuple[int, int],
        probability_postfix: str = "_p_predicted",
        prediction_postfix: str = "_predicted",
    ):
        self._classes = classes
        self._logistic_transform = logistic_transform
        self._input_shape = input_shape
        self._probability_postfix = probability_postfix
        self._prediction_postfix = prediction_postfix

        # init session
        self._session = onnxruntime.InferenceSession(file)

    def predict(
        self,
        images: ImageOrSequence,
        norm_mean: float = None,
        norm_std: float = None,
        thresh: float = 0.5,
    ) -> ImageOrSequence:

        if (norm_mean is None or norm_std is None) and len(images) < 20:
            raise RuntimeError(
                "Please provide a sufficient amount of images to automatically compute statistics"
            )
        elif norm_mean is None or norm_std is None:
            # automatically compute statistics using all images
            imgs = np.stack(
                [images[i].data.astype(np.float) for i in range(len(images))]
            )
            norm_mean = np.mean(imgs)
            norm_std = np.std(imgs)

        def apply(x: Image):
            data = x.data.astype(np.float)

            # conditionally resize
            if self._input_shape is not None:
                tgt = [
                    s1 / s2 for s1, s2 in zip(list(self._input_shape), list(data.shape))
                ]
                data = rescale(data, scale=tgt)

            # normalize
            data = (data - norm_mean) / norm_std

            # predict
            data = np.tile(data, (1, 3, 1, 1)).astype(np.float32)
            ort_inputs = {self._session.get_inputs()[0].name: data}
            pred = self._session.run(None, ort_inputs)[0].flatten()

            # sigmoid
            if self._logistic_transform:
                pred = special.expit(pred)

            d1 = {
                k + self._probability_postfix: pred[i]
                for k, i in zip(self._classes, range(len(self._classes)))
            }
            d2 = {
                k + self._prediction_postfix: pred[i] > thresh
                for k, i in zip(self._classes, range(len(self._classes)))
            }
            d1.update(d2)
            return d1

        return images.meta_from_fn(apply, progress_bar=True)



"""Read and write images"""

import json
import logging
import pickle
import urllib.parse
from datetime import date, datetime
from functools import reduce
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.lib.npyio import save
from pvinspect.common import util
from pvinspect.common.types import ObjectAnnotations, PathOrStr
from pvinspect.data.image import Modality
from pyparsing import ParseException
from shapely.errors import WKTReadingError
from shapely.geometry import Point, Polygon
from shapely.wkt import dumps as shapely_dumps
from shapely.wkt import loads as shapely_loads
from skimage import color, img_as_float, img_as_uint, io
from tqdm.auto import tqdm

from .exceptions import UnsupportedModalityException
from .image import *


def _prepare_json_meta(meta):

    # numpy scalar -> builtin
    if type(meta).__module__ == np.__name__ and meta.ndim == 0:
        meta = np.asscalar(meta)

    if isinstance(meta, dict):
        return {k: _prepare_json_meta(v) for k, v in meta.items()}
    elif isinstance(meta, list):
        return [_prepare_json_meta(v) for v in meta]
    elif isinstance(meta, (Polygon, Point)):
        return shapely_dumps(meta)
    elif isinstance(meta, (date, datetime)):
        return meta.isoformat()
    elif isinstance(meta, Modality):
        return str(meta)
    elif isinstance(meta, (str, float, int, bool)):
        return meta
    else:
        return None


def _load_json_meta_hook(pairs):
    result = dict()
    for k, v in pairs.items():
        if isinstance(v, str):
            if v == "Modality.EL_IMAGE":
                result[k] = Modality.EL_IMAGE
                continue
            if v == "Modality.PL_IMAGE":
                result[k] = Modality.PL_IMAGE
                continue

            try:
                result[k] = datetime.fromisoformat(v)
                continue
            except ValueError:
                result[k] = v
            try:
                with util.suppress_stderr():
                    result[k] = shapely_loads(v)
                continue
            except WKTReadingError:
                result[k] = v
        else:
            result[k] = v
    return result


def __assurePath(p: PathOrStr) -> Path:
    if isinstance(p, str):
        return Path(p)
    else:
        return p


def _get_meta_path(img_path: Path) -> Path:
    return img_path.parent / "{}.json".format(img_path.name)


def _read_image(
    path: PathOrStr,
    is_module_image: bool,
    is_partial_module: bool,
    lazy: bool,
    modality: int = None,
    cols: Optional[int] = None,
    rows: Optional[int] = None,
    first_col: Optional[int] = None,
    first_row: Optional[int] = None,
    force_dtype: Union[DType, None] = None,
    meta=None,
):
    assert (
        is_module_image
        and not is_partial_module
        or is_partial_module
        and not is_module_image
        or not is_partial_module
        and not is_module_image
    )

    path = __assurePath(path)

    def _read(path):
        img = io.imread(path)

        if img.dtype == ">u2":
            # big endian -> little endian
            img = img.astype(np.uint16)

        if img.ndim == 3 and (img.dtype == np.float32 or img.dtype == np.float64):
            img = color.rgb2gray(img)
        elif img.ndim == 3:
            img = img_as_uint(color.rgb2gray(img))

        if force_dtype == DType.FLOAT:
            img = img.astype(DTYPE_FLOAT)
        elif force_dtype == DType.UNSIGNED_INT:
            img = img.astype(DTYPE_UNSIGNED_INT)
        elif force_dtype == DType.INT:
            img = img.astype(DTYPE_INT)

        if (img.dtype == np.float32 or img.dtype == np.float64) and (
            img.min() < 0.0 or img.max() > 1.0
        ):
            raise RuntimeWarning(
                'Image "{}" is of type float but not scaled between 0 and 1. This might cause trouble later. You might want to force conversion to another datatype using force_dtype=DType.UNSIGNED_INT (for example).'.format(
                    path
                )
            )
        return img

    # conditionally enable lazy loading
    img = _read(path) if not lazy else Image.LazyData(partial(_read, path))

    # try to read meta file
    meta_path = _get_meta_path(path)
    meta_series: pd.Series = None
    if meta is not None:
        meta_series = meta
    elif meta_path.is_file():
        with open(meta_path, "r") as f:
            meta_series = pd.Series(json.load(f, object_hook=_load_json_meta_hook))

    # merge modality, rows and cols
    if (
        meta_series is not None
        and "modality" in meta_series.keys()
        and modality is None
    ):
        modality = meta_series["modality"]
    if meta_series is not None and "rows" in meta_series.keys() and rows is None:
        rows = meta_series["rows"]
    if meta_series is not None and "cols" in meta_series.keys() and cols is None:
        cols = meta_series["cols"]

    if is_partial_module:
        return PartialModuleImage(
            img,
            modality,
            path,
            cols,
            rows,
            first_row=first_row,
            first_col=first_col,
            meta=meta,
        )
    elif is_module_image:
        return ModuleImage(img, modality, path, cols, rows, meta=meta_series)
    else:
        return Image(img, path, modality, meta=meta_series)


def _get_meta_cache_path(sequence_path: Path) -> Path:
    return sequence_path / ".cached_meta.pck"


def _read_images(
    path: PathOrStr,
    is_module_image: bool,
    is_partial_module: bool,
    lazy: bool,
    same_camera: bool = False,
    modality: int = None,
    cols: int = None,
    rows: int = None,
    N: int = 0,
    pattern: Union[str, Tuple[str]] = ("*.png", "*.tif", "*.tiff", "*.bmp"),
    allow_different_dtypes=False,
    force_dtype: DType = None,
) -> ImageSequence:
    path = __assurePath(path)

    if isinstance(pattern, str):
        pattern = [pattern]

    # find files and skip if more than N
    imgpaths: List[Path] = list(
        reduce(lambda x, y: x + y, [list(path.glob(pat)) for pat in pattern])
    )
    imgpaths.sort()
    if N > 0 and N < len(imgpaths):
        imgpaths = imgpaths[:N]

    # check if cached meta data is available
    cache_file = _get_meta_cache_path(path)
    cached_meta: pd.DataFrame = None
    if cache_file.is_file():
        with open(cache_file, "rb") as f:
            try:
                cached_meta = pd.read_pickle(f)
            except:
                cache_file.unlink()

    imgs = list()
    missing_meta: Dict[str, pd.Series] = dict()
    for fn in tqdm(imgpaths):
        if cached_meta is not None and fn.name in cached_meta.keys():
            img = _read_image(
                fn,
                is_module_image,
                is_partial_module,
                lazy,
                modality,
                cols,
                rows,
                force_dtype=force_dtype,
                meta=cached_meta.loc[fn.name],
            )
        else:
            img = _read_image(
                fn,
                is_module_image,
                is_partial_module,
                lazy,
                modality,
                cols,
                rows,
                force_dtype=force_dtype,
            )
            missing_meta[fn.name] = img.meta_to_pandas()
        imgs.append(img)

    # update/create meta cache file
    if len(missing_meta.keys()) > 0:
        missing_meta_df = pd.DataFrame(missing_meta).T
        new_meta = (
            pd.concat([cached_meta, missing_meta_df])
            if cached_meta is not None
            else missing_meta_df
        )
        new_meta.to_pickle(cache_file)

    # TODO: How to handle this with lazy loading?
    # if not same_camera:
    #    homogeneous_types = np.all(
    #        np.array([img.dtype == imgs[0].dtype for img in imgs])
    #    )
    #    shapes = [img.shape for img in imgs]
    #    homogeneous_shapes = np.all(np.array([s == shapes[0] for s in shapes]))
    #    target_shape = np.max(shapes, axis=0)

    #    if not homogeneous_shapes:
    #        logging.warning(
    #            "The original images are of different shape. They might not be suited for all applications (for example superresolution)."
    #        )

    #    if not homogeneous_types:
    #        # target type is determined by the first image
    #        logging.warning(
    #            "The original images are of different type. They are converted to the type of the first image ({})".format(
    #                imgs[0].dtype
    #            )
    #        )
    #        conv = (
    #            img_as_float
    #            if imgs[0].dtype == np.float32 or imgs[0].dtype == np.float64
    #            else img_as_uint
    #        )
    #        imgs = [conv(img) for img in imgs]

    if is_module_image or is_partial_module:
        res = ModuleImageSequence(
            imgs, same_camera=same_camera, allow_different_dtypes=allow_different_dtypes
        )
    else:
        res = ImageSequence(imgs, same_camera, allow_different_dtypes)

    return res


def read_image(
    path: PathOrStr, modality: int = None, force_dtype: DType = None, lazy: bool = True,
) -> Image:
    """Read a single image of a solar module and return it

    Args:
        path (PathOrStr): Path to the file to be read
        modality (int): The imaging modality
        force_dtype (DType): Force images to have this datatype
        lazy (bool): Enable lazy loading for this image

    Returns:
        image: The module image
    """

    return _read_image(
        path=path,
        is_module_image=False,
        is_partial_module=False,
        lazy=lazy,
        modality=modality,
        force_dtype=force_dtype,
    )


def read_images(
    path: PathOrStr,
    same_camera: bool = False,
    modality: int = None,
    N: int = 0,
    pattern: Union[str, Tuple[str]] = ("*.png", "*.tif", "*.tiff", "*.bmp"),
    allow_different_dtypes=False,
    force_dtype: DType = None,
    lazy: bool = True,
) -> ImageSequence:
    """Read a sequence of images and return it

    Args:
        path (PathOrStr): Path to the sequence
        same_camera (bool): Indicate, if all images are from the same camera and hence share the same intrinsic parameters
        modality (int): The imaging modality
        N (int): Only read first N images
        pattern (Union[str, Tuple[str]]): Files must match any of the given pattern
        allow_different_dtypes (bool): Allow images to have different datatypes?
        force_dtype (DType): Force images to have this datatype
        lazy (bool): Enable lazy loading for these images

    Returns:
        image: The image sequence
    """

    return _read_images(
        path,
        is_module_image=False,
        same_camera=same_camera,
        is_partial_module=False,
        lazy=lazy,
        modality=modality,
        pattern=pattern,
        allow_different_dtypes=allow_different_dtypes,
        N=N,
        force_dtype=force_dtype,
    )


def read_module_image(
    path: PathOrStr,
    modality: Modality,
    cols: int = None,
    rows: int = None,
    lazy: bool = True,
) -> ModuleImage:
    """Read a single image of a solar module and return it

    Args:
        path (PathOrStr): Path to the file to be read
        modality (int): The imaging modality
        cols (int): Number of columns of cells
        rows (int): Number of rows of cells
        lazy (bool): Enable lazy loading for this image

    Returns:
        image: The module image
    """

    return _read_image(
        path=path,
        is_module_image=True,
        is_partial_module=False,
        lazy=lazy,
        modality=modality,
        cols=cols,
        rows=rows,
    )


def read_module_images(
    path: PathOrStr,
    modality: int,
    same_camera: bool = False,
    cols: int = None,
    rows: int = None,
    N: int = 0,
    pattern: Union[str, Tuple[str]] = ("*.png", "*.tif", "*.tiff", "*.bmp"),
    allow_different_dtypes=False,
    force_dtype: DType = None,
    lazy: bool = True,
) -> ModuleImageSequence:
    """Read a sequence of module images and return it

    Args:
        path (PathOrStr): Path to the sequence
        modality (int): The imaging modality
        same_camera (bool): Indicate, if all images are from the same camera and hence share the same intrinsic parameters
        cols (int): Number of columns of cells
        rows (int): Number of rows of cells
        N (int): Only read first N images
        pattern (Union[str, Tuple[str]]): Files must match any of the given pattern
        allow_different_dtypes (bool): Allow images to have different datatypes?
        force_dtype (DType): Force images to have this datatype
        lazy (bool): Enable lazy loading for these images

    Returns:
        image: The module image sequence
    """

    return _read_images(
        path=path,
        modality=modality,
        same_camera=same_camera,
        is_partial_module=False,
        is_module_image=True,
        lazy=lazy,
        cols=cols,
        rows=rows,
        N=N,
        pattern=pattern,
        allow_different_dtypes=allow_different_dtypes,
        force_dtype=force_dtype,
    )


def read_partial_module_image(
    path: PathOrStr,
    modality: int,
    cols: Optional[int] = None,
    rows: Optional[int] = None,
    first_col: Optional[int] = None,
    first_row: Optional[int] = None,
    lazy: bool = True,
) -> ModuleImage:
    """Read a single partial view of a solar module and return it

    Args:
        path (PathOrStr): Path to the file to be read
        modality (int): The imaging modality
        cols (int): Number of completely visible columns of cells
        rows (int): Number of completely visible rows of cells
        lazy (bool): Enable lazy loading for this image

    Returns:
        image: The module image
    """

    return _read_image(
        path=path,
        is_module_image=False,
        is_partial_module=True,
        lazy=lazy,
        modality=modality,
        cols=cols,
        rows=rows,
        first_col=first_col,
        first_row=first_row,
    )


def read_partial_module_images(
    path: PathOrStr,
    modality: int,
    same_camera: bool = False,
    cols: int = None,
    rows: int = None,
    N: int = 0,
    pattern: Union[str, Tuple[str]] = ("*.png", "*.tif", "*.tiff", "*.bmp"),
    allow_different_dtypes=False,
    force_dtype: DType = None,
    lazy: bool = True,
) -> ModuleImageSequence:
    """Read a sequence of partial views of solar modules and return it

    Args:
        path (PathOrStr): Path to the sequence
        modality (int): The imaging modality
        same_camera (bool): Indicate, if all images are from the same camera and hence share the same intrinsic parameters
        cols (int): Number of completely visible columns of cells
        rows (int): Number of completely visible rows of cells
        N (int): Only read first N images
        pattern (Union[str, Tuple[str]]): Files must match any of the given pattern
        allow_different_dtypes (bool): Allow images to have different datatypes?
        force_dtype (DType): Force images to have this datatype
        lazy (bool): Enable lazy loading for these images

    Returns:
        image: The module image sequence
    """

    return _read_images(
        path=path,
        is_module_image=False,
        is_partial_module=True,
        same_camera=same_camera,
        lazy=lazy,
        modality=modality,
        cols=cols,
        rows=rows,
        N=N,
        pattern=pattern,
        allow_different_dtypes=allow_different_dtypes,
        force_dtype=force_dtype,
    )


def save_image(
    filename: PathOrStr,
    image: Image,
    with_visusalization: bool = False,
    save_meta: bool = False,
    **kwargs
):
    """Write an image to disk. Float64 is automatically converted to float32 in order to be compatible to ImageJ.

    Args:
        filename (PathOrStr): Filename of the resulting image
        image (Image): The image
        with_visualization (bool): Include the same visualizations as with image.show() or sequence.head()
        save_meta (bool): Save meta data to separate json file (cannot be used with visualization)
    """
    filename = __assurePath(filename)

    if with_visusalization and save_meta:
        logging.error("Cannot save meta data for image with visualization")
        return

    if with_visusalization:
        plt.clf()
        image.show(**kwargs)
        plt.savefig(filename, **kwargs)
    else:
        if image.dtype == DType.FLOAT:
            io.imsave(filename, image.data.astype(np.float32), check_contrast=False)
        else:
            io.imsave(filename, image.data, check_contrast=False)
        if save_meta and len(image.list_meta()) > 0:
            meta = {k: image.get_meta(k) for k in image.list_meta()}
            meta = _prepare_json_meta(meta)
            meta_path = _get_meta_path(filename)
            with open(meta_path, "w") as f:
                json.dump(meta, f)

    # delete cached entry
    cache_file = _get_meta_cache_path(filename.parent)
    if save_meta and cache_file.is_file():
        cached_meta: pd.DataFrame = pd.read_pickle(cache_file)
        cached_meta.drop(filename.name, inplace=True)
        cached_meta.to_pickle(cache_file)


def save_images(
    path: PathOrStr,
    sequence: ImageSequence,
    mkdir: bool = True,
    with_visualization: bool = False,
    hierarchical: List[str] = None,
    include_meta_keys: bool = True,
    save_meta: bool = False,
    filename_prefix: str = "",
    filename_suffix: str = "",
    filename_hook: Callable[[Image], Path] = lambda image: Path(image.path.name),
    **kwargs
):
    """Write a sequence of images to disk

    Args:
        path (PathOrStr): Target directory
        sequence (ImageSequence): The sequence of images
        mkdir (bool): Automatically create missing directories
        with_visualization (bool): Include the same visualizations as with image.show() or sequence.head()
        hierarchical (List[str]): Create a directory hierarchy using given meta keys
        include_meta_keys (bool): Indicate, if meta keys should be included in the folder names
        save_meta (bool): Save image meta in json file
        filename_prefix (str): Prepend every filename with this
        filename_suffix (str): Add this suffix to every filename
        filename_hook (Callable[[Image], Path]): Callable that is used to determine the filename
    """

    path = __assurePath(path)

    if mkdir:
        path.mkdir(parents=True, exist_ok=True)

    # check if meta cache exists and delete if necessary
    cache_file = _get_meta_cache_path(path)
    if save_meta and cache_file.is_file():
        cache_file.unlink()

    for image in tqdm(sequence.images):

        # determine filename
        fn = filename_hook(image)
        fn = Path(filename_prefix + fn.stem + filename_suffix + fn.suffix)

        fpath = path
        if hierarchical is not None:
            for mk in hierarchical:
                if not include_meta_keys:
                    fpath /= str(image.get_meta(mk))
                else:
                    fpath /= "{}_{}".format(mk, str(image.get_meta(mk)))

        if mkdir:
            fpath.mkdir(parents=True, exist_ok=True)

        fpath /= fn
        save_image(
            fpath,
            image,
            with_visusalization=with_visualization,
            save_meta=save_meta,
            **kwargs
        )


def load_json_object_masks(path: PathOrStr) -> ObjectAnnotations:
    """Load object annotations from file

    Args:
        path (PathOrStr): Path to the annotations file
    
    Returns:
        annotations: Dict with filenames a keys and a list of annotations, where every
            annotation is a tuple of "classname" and a Polygon
    """

    path = __assurePath(path)

    with open(path, "r") as f:
        js = json.load(f)

    if isinstance(js, list):
        result = dict()
        for item in js:
            anns = list()
            for k, v in item["Label"].items():
                for ann in v:
                    poly = Polygon([(x["x"], x["y"]) for x in ann["geometry"]])
                    anns.append((k, poly))
            result[item["External ID"]] = anns
        return result
    elif isinstance(js, dict):
        result = dict()
        prefix_len = 124

        # id -> category name
        catbyid = dict()
        for item in js["categories"]:
            catbyid[item["id"]] = item["name"]

        for img in js["images"]:
            fn = urllib.parse.unquote(img["file_name"][prefix_len:])
            result[fn] = list()
            for item in js["annotations"]:
                if item["image_id"] == img["id"]:
                    x = item["segmentation"][0]
                    poly = Polygon(
                        [(x[2 * i + 0], x[2 * i + 1]) for i in range(len(x) // 2)]
                    )
                    result[fn].append((catbyid[item["category_id"]], poly))

        return result



import numpy as np
from pvinspect.data import ModuleImage, ModuleImageSequence
from pvinspect.data.image import Image, ImageSequence, Modality
from pvinspect.preproc.detection import locate_module_and_cells, segment_module_part


def _stitch_two(
    img1: Image, img2: Image, cell_size: int, overlap: int, direction: str
) -> Image:
    if direction == "hor":
        img1 = segment_module_part(
            img1,
            0,
            0,
            img1.get_meta("cols") - (overlap // 2),
            img1.get_meta("rows"),
            cell_size,
        )
        img2 = segment_module_part(
            img2,
            overlap // 2,
            0,
            img2.get_meta("cols") - (overlap // 2),
            img2.get_meta("rows"),
            cell_size,
        )
        return img1.from_self(
            data=np.concatenate([img1.data, img2.data], axis=1),
            meta=dict(cols=img1.get_meta("cols") + img2.get_meta("cols") - overlap),
        )
    else:
        img1 = segment_module_part(
            img1,
            0,
            0,
            img1.get_meta("cols"),
            img1.get_meta("rows") - (overlap // 2),
            cell_size,
        )
        img2 = segment_module_part(
            img2,
            0,
            overlap // 2,
            img2.get_meta("cols"),
            img2.get_meta("rows") - (overlap // 2),
            cell_size,
        )
        return img1.from_self(
            data=np.concatenate([img1.data, img2.data], axis=0),
            meta=dict(rows=img1.get_meta("rows") + img2.get_meta("rows") - overlap),
        )


def _scale_median_brightness(x: np.ndarray, target: float) -> np.ndarray:
    source = np.median(x)
    return (x * (target / source)).astype(x.dtype)


def stitch_modules(
    images: ImageSequence,
    n_horizontal: int = 1,
    n_vertical: int = 1,
    overlap_horizontal: int = 0,
    overlap_vertical: int = 0,
    equalize_intensity: bool = False,
) -> Image:
    """Locate and stitch partial recordings of a module

    Args:
        images (ImageSequence): Partial recordings to be stitched together (needs to be order left -> right / top -> bottom)
        n_horizontal (int): Number of partial recordings in horizontal direction
        n_vertical (int): Number of partial recordings in vertical direction
        overlap_horizontal (int): Number of fully visible cells that overlap between any two partial recordings
            in horizontal direction
        overlap_vertical (int): Number of fully visible cells that overlap between any two partial recordings
            in vertical direction
        equalize_intenity (bool): Match the median intensity of every partial recording to the median intensity of the first

    Returns:
        image (Image): The stitched image
    """

    if n_horizontal < 1 or n_vertical < 1:
        raise RuntimeError(
            "Number of images for stitching must not be smaller than 1 in any direction"
        )

    if n_horizontal > 2 or n_vertical > 2:
        raise NotImplementedError(
            "Stitching is only implemented for configurations 1x2, 2x1 and 2x2"
        )

    if overlap_horizontal < 0 or overlap_vertical < 0:
        raise RuntimeError("Invalid overlap")

    if overlap_horizontal % 2 != 0 or overlap_vertical % 2 != 0:
        raise NotImplementedError("Stitching is only implemented for even overlap")

    # determine spacing based on first image
    t = images[0].get_meta("transform")
    cell_size = np.linalg.norm(
        t(np.array([[1.0, 1.0]]))[0] - t(np.array([[0.0, 0.0]]))[0]
    )

    # equalize brightness if set
    if equalize_intensity:
        ref = np.median(images[0].data)  # use the first image as reference
        images = images.apply_image_data(_scale_median_brightness, target=ref)

    # stitch
    if n_horizontal == 1 and n_vertical == 1:
        return images[0]
    elif n_horizontal == 2 and n_vertical == 1:
        return _stitch_two(images[0], images[1], cell_size, overlap_horizontal, "hor")
    elif n_horizontal == 1 and n_vertical == 2:
        return _stitch_two(images[0], images[1], cell_size, overlap_vertical, "ver")
    else:  # n_horizontal == 2 and n_vertical == 2
        row1 = _stitch_two(images[0], images[1], cell_size, overlap_horizontal, "hor")
        row2 = _stitch_two(images[2], images[3], cell_size, overlap_horizontal, "hor")
        return row1.from_self(
            data=np.concatenate([row1.data, row2.data], axis=0),
            meta=dict(rows=row1.get_meta("rows") + row2.get_meta("rows")),
        )


# def locate_and_stitch_modules(images: ImageSequence, n_horizontal: int = 1, )
def locate_and_stitch_modules(
    images: ImageSequence,
    rows: int,
    cols: int,
    n_horizontal: int = 1,
    n_vertical: int = 1,
    overlap_horizontal: int = 0,
    overlap_vertical: int = 0,
    equalize_intensity: bool = False,
) -> Image:
    """Locate and stitch partial recordings of a module

    This method applies localization of modules followed by stitching. It relies on an exact specification
    of the visible module geometry (by means of rows, cols, n_horizontal, n_vertical, overlap_horizontal,
    overlap_vertical). Furthermore images need to be given in the specified order.

    This method optionally provides adaptation of intensities of the partial recordings. This is turned off
    by default, since it changes the original intensities, which might be undesirable in some cases.

    Args:
        images (ImageSequence): Partial recordings to be stitched together (needs to be order left -> right / top -> bottom)
        rows (int): Number of fully visible rows of cells in every partial recording
        cols (int): Number of fully visible columns of cells in every partial recording
        n_horizontal (int): Number of partial recordings in horizontal direction
        n_vertical (int): Number of partial recordings in vertical direction
        overlap_horizontal (int): Number of fully visible cells that overlap between any two partial recordings
            in horizontal direction
        overlap_vertical (int): Number of fully visible cells that overlap between any two partial recordings
            in vertical direction
        equalize_intensity (bool): Match the median intensity of every partial recording to the median intensity of the first

    Returns:
        image (Image): The stitched image
    """

    modimages = ModuleImageSequence(
        [
            ModuleImage(x.data, modality=Modality.EL_IMAGE, cols=cols, rows=rows)
            for x in images
        ]
    )

    # locate
    modimages = locate_module_and_cells(modimages, orientation="horizontal")

    # stitch
    return stitch_modules(
        modimages,
        n_horizontal,
        n_vertical,
        overlap_horizontal,
        overlap_vertical,
        equalize_intensity,
    )



"""
Allows exit() to work if script is invoked with IPython without
raising NameError Exception. Keeps kernel alive.

Use: import variable 'exit' in target script with 'from ipython_exit import exit'    
"""

import sys
from io import StringIO


class IpyExit(SystemExit):
    """Exit Exception for IPython.

    Exception temporarily redirects stderr to buffer.
    """

    def __init__(self):
        # print("exiting")  # optionally print some message to stdout, too
        # ... or do other stuff before exit
        sys.stderr = StringIO()

    def __del__(self):
        sys.stderr.close()
        sys.stderr = sys.__stderr__  # restore from backup


def ipy_exit():
    raise IpyExit


try:
    __IPYTHON__
    exit = ipy_exit  # rebind to custom exit
except:
    exit = exit  # just make exit importable



# preprocessing
PREPROC_LOWER_PERCENTILE = 10
PREPROC_UPPER_PERCENTILE = 80

# module/cell detection
GAUSSIAN_RELATIVE_SIGMA = 0.01
RANSAC_THRES = 0.1  # 10% of cell size
OUTER_CORNER_THRESH_FACTOR = 2  # gradient must be this times the std dev of gradient to be accepted as outer corner
CORNER_DETECTION_PATCH_SIZE = 1  # size of warped fragment
CROP_MARGIN = 0.5  # margin around cropped modules (relative to cell size)
MODULE_DETECTION_PEAK_THRESH = 1  # peak threshold relative to mean signal
CORNER_DETECTION_MAX_PEAKS = 5  # maximum number of peaks in corner detection
CORNER_DETECTION_PEAK_THRESH = (
    1  # peak threhould for corner detection relative to mean signal
)

# outlier rejection
REPROJ_THRES = 0.1  # threshold on reprojection error to reject images; 5% of cell size
USE_RANSAC = False  # apply ransac during joint parameter estimation



import numpy as np
from pvinspect.common import transform as t
from pvinspect.common import util
from test.utilities import assert_equal


def test_distort_undistort():
    x = np.random.rand(1000, 2) - 0.5
    k = [0.1]

    xt = t.distort(x, k)
    xtt = t.distort_inverse(xt, k)
    err = np.linalg.norm(x - xtt)
    assert err <= 1e-8

    k[-1] *= -1
    xt = t.distort(x, k)
    xtt = t.distort_inverse(xt, k)
    err = np.linalg.norm(x - xtt)
    assert err <= 1e-8


def test_find_homography_exact():
    H = np.array([[0.8, 0, 1.2], [0.2, 2.0, 10], [0, 0, 1]])

    src = np.random.rand(4, 3)
    src[:, -1] = 1
    dest = H.dot(src.T).T
    dest[:, :2] /= dest[:, 2].reshape(-1, 1)

    H2, residual = t.find_homography(src[:, :2], dest[:, :2], True)
    dest2 = H2.dot(src.T).T
    dest2 /= dest2[:, 2].reshape(-1, 1)
    err = np.linalg.norm((dest - dest2).flatten())

    assert residual < 0.1
    assert err < 1e-8
    assert np.linalg.norm(H.flatten() - H2.flatten()) < 1e-8


def test_find_homography_l2():
    H = np.array([[0.8, 0, 1.2], [0.2, 2.0, 10], [0, 0, 1]])

    src = np.random.rand(100, 3)
    src[:, -1] = 1
    dest = H.dot(src.T).T
    dest[:, :2] /= dest[:, 2].reshape(-1, 1)
    err_term = np.random.uniform(-0.2, 0.2, dest[:, :2].shape)
    dest[:, :2] += err_term

    H2, residual = t.find_homography(src[:, :2], dest[:, :2], True)
    dest2 = H2.dot(src.T).T
    dest2 /= dest2[:, 2].reshape(-1, 1)
    err = np.mean(np.linalg.norm(dest - dest2, ord=2, axis=1))
    err_term_norm = np.mean(np.linalg.norm(err_term, ord=2, axis=1))

    assert np.abs(err - err_term_norm) < 0.05
    assert np.abs(err - residual) < 0.5


def test_apply_homography():
    H = np.array([[0.8, 0, 1.2], [0.2, 2.0, 10], [0, 0, 1]])
    src = np.random.rand(1000, 3)
    src[:, -1] = 1
    dest = H.dot(src.T).T
    dest[:, :2] /= dest[:, 2].reshape(-1, 1)

    res = t.apply_homography(src[:, :2].reshape(10, 100, 2), H)
    res = res.reshape(1000, 2)
    err = np.mean(np.linalg.norm(dest[:, :2] - res[:, :2], axis=1))

    assert err < 1e-8


def test_find_homography_ransac():
    H = np.array([[0.8, 0, 1.2], [0.2, 2.0, 10], [0, 0, 1]])

    src = np.random.rand(10, 3)
    src[:, -1] = 1
    dest = H.dot(src.T).T
    dest[:, :2] /= dest[:, 2].reshape(-1, 1)
    dest[:2] += 1  # 2 outlier

    H2, is_inlier, N = t.find_homography_ransac(src[:, :2], dest[:, :2], 0.01)
    dest2 = H2.dot(src.T).T
    dest2 /= dest2[:, 2].reshape(-1, 1)
    err = np.linalg.norm((dest[3:] - dest2[3:]).flatten())

    assert err < 1e-8
    assert np.sum(is_inlier) == 8
    assert not is_inlier[0]
    assert not is_inlier[1]


# def test_full_model():
#    N_imgs = 5
#
#    Rt = [util.random_rotation_translation_matrix_3d() for i in range(N_imgs)]
#    dist = [0.05]
#    A = np.array([[900, 0.00, 255], [0, 1200, 255], [0, 0, 1]])
#    src = [np.random.rand(40, 2) * 10 for i in range(N_imgs)]
#    dest = [
#        t.apply_homography(t.distort(t.apply_homography(x, y[:, [0, 1, 3]]), dist), A)
#        for x, y in zip(src, Rt)
#    ]
#
#    Rt2, A2, dist2 = t.find_full_model(src, dest, 1, image_width=500, image_height=500)
#
#    assert np.mean(np.abs(A - A2)) < 2
#    assert_equal(dist[0], dist2[0], 1e-1)
#    for i in range(N_imgs):
#        assert np.mean(np.abs(Rt[i] - Rt2[i])) < 2


def test_homography_transform():
    H = np.array([[20, 0, 10], [0, 30, 15], [0, 0, 1]]).astype(np.float64)
    src = np.random.rand(20, 2)
    dest = t.apply_homography(src, H)

    transform = t.HomographyTransform(src, dest)
    src_t = transform(src)
    src_res = transform.inv()(src_t)

    err = np.mean(np.linalg.norm(src - src_res, axis=1))
    assert err < 1e-8


def test_full_transform():
    A = np.array([[1200, 0, 500], [0, 900, 450], [0, 0, 1]])
    Rt = util.random_rotation_translation_matrix_3d()
    dist = [0.5]

    src = np.random.rand(20, 2)
    dest = t.apply_pinhole(src, Rt, A, dist)

    transform = t.FullTransform(src, dest, 1000, 900, n_dist_coeff=1)
    src_t = transform(src)
    src_res = transform.inv()(src_t)

    err = np.mean(np.linalg.norm(src - src_res, axis=1))
    assert err < 1e-2

    err = np.mean(np.linalg.norm(src_t - dest, axis=1))
    assert err < 1e-2


def test_full_multi_transform():
    A = np.array([[1200, 0, 500], [0, 900, 450], [0, 0, 1]])
    Rt = [util.random_rotation_translation_matrix_3d() for i in range(5)]
    dist = [0.15]
    src = [np.random.rand(20, 2) for i in range(5)]
    dest = [t.apply_pinhole(s, r, A, dist) for s, r in zip(src, Rt)]

    transform = t.FullMultiTransform(src, dest, 1000, 900, ransac=False, n_dist_coeff=1)
    src_t = [t(x) for t, x in zip(transform, src)]
    src_res = [t.inv()(x) for t, x in zip(transform, src_t)]

    errs = [np.mean(np.linalg.norm(x - y, axis=1)) for x, y in zip(src_res, src)]
    assert np.mean(errs) < 1e-4

    transform = transform.inv()
    src_res = [t(x) for t, x in zip(transform, src_t)]

    errs = [np.mean(np.linalg.norm(x - y, axis=1)) for x, y in zip(src_res, src)]
    assert np.mean(errs) < 1e-4


def test_distortion_model_is_consistent():
    k1 = 0.08
    k2 = -0.01
    src = np.random.rand(20, 2)
    r = np.linalg.norm(src, axis=1).reshape(-1, 1)
    src_dist = src + src * k1 * r ** 2 + src * k2 * r ** 4
    src_dist2 = t.distort(src, [k1, k2])

    err = np.mean(np.abs(src_dist - src_dist2))
    assert err < 1e-8



from __future__ import annotations

"""Provides classes to store and visualize images with metadata"""

import copy
import inspect
import logging
import math
import re
import sys
from enum import Enum
from functools import lru_cache, partial, wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pvinspect.common._ipy_exit import exit
from pvinspect.common.transform import Transform
from skimage import img_as_float64, img_as_int, img_as_uint, io
from tqdm.autonotebook import tqdm

# this is a pointer to the module object instance itself
this = sys.modules[__name__]


# modality
class Modality(Enum):
    EL_IMAGE = (0,)
    PL_IMAGE = 1


EL_IMAGE = Modality.EL_IMAGE
"""Indicate an electroluminescense (EL) image"""
PL_IMAGE = Modality.PL_IMAGE
"""Indicate a photoluminescense (PL) image"""


# datatypes
DTYPE_INT = np.int32
DTYPE_UNSIGNED_INT = np.uint16
DTYPE_UNSIGNED_BYTE = np.uint8
DTYPE_FLOAT = np.float64
img_as_float = img_as_float64

# caching
SEQUENCE_MAX_CACHE_SIZE = 5000


class DType(Enum):
    INT = 0
    UNSIGNED_INT = 1
    FLOAT = 2
    UNSIGNED_BYTE = 3


# global list of plugins that are called on every .show()
this.show_plugins = list()


def register_show_plugin(callable, priority: int = 0):
    """Register a new plugin that is called on every .show()

    Args:
        callable: Callable that receives the image as a first argument and variable arguments to .show() next
        priority (int): Plugins are invoked in the order of increasing priority (highest priority is invoked
            last and hence appears on top)
    """
    this.show_plugins.append((priority, callable))
    this.show_plugins = sorted(this.show_plugins, key=lambda x: x[0])


def _invoke_show_plugins(image, **kwargs):
    for p in this.show_plugins:
        p[1](image, **kwargs)


def _register_default_plugins():
    def show_cell_crossings(
        image: ModuleImage, show_cell_crossings: bool = True, **kwargs
    ):
        if (
            show_cell_crossings
            and isinstance(image, ModuleImage)
            and image.has_meta("transform")
        ):
            grid = image.grid()
            coords = image.get_meta("transform").__call__(grid)
            plt.scatter(coords[:, 0], coords[:, 1], c="yellow", marker="+")

    register_show_plugin(show_cell_crossings)

    def multimodule_show_boxes(
        image: Image,
        multimodule_show_boxes: bool = True,
        multimodule_highlight_selection: bool = True,
        multimodule_boxes_linewidth: int = 2,
        **kwargs
    ):
        if (
            multimodule_show_boxes
            and isinstance(image, Image)
            and image.has_meta("multimodule_boxes")
        ):
            for i, box in enumerate(image.get_meta("multimodule_boxes")):
                color = (
                    "red"
                    if i == image.get_meta("multimodule_index")
                    and multimodule_highlight_selection
                    else "yellow"
                )
                plt.plot(
                    *box[1].exterior.xy,
                    linewidth=multimodule_boxes_linewidth,
                    color=color,
                )

    register_show_plugin(multimodule_show_boxes)

    def multimodule_show_numbers(
        image: Image,
        multimodule_show_numbers: bool = True,
        multimodule_highlight_selection: bool = True,
        multimodule_numbers_fontsize: int = 20,
        **kwargs
    ):
        if (
            multimodule_show_numbers
            and isinstance(image, Image)
            and image.has_meta("multimodule_boxes")
        ):
            for i, box in enumerate(image.get_meta("multimodule_boxes")):
                bgcolor = (
                    "red"
                    if i == image.get_meta("multimodule_index")
                    and multimodule_highlight_selection
                    else "white"
                )
                textcolor = (
                    "white"
                    if i == image.get_meta("multimodule_index")
                    and multimodule_highlight_selection
                    else "black"
                )
                plt.text(
                    box[1].centroid.x,
                    box[1].centroid.y,
                    s=str(i),
                    color=textcolor,
                    fontsize=multimodule_numbers_fontsize,
                    bbox=dict(facecolor=bgcolor, alpha=0.8),
                    ha="center",
                    va="center",
                )

    register_show_plugin(multimodule_show_numbers)

    def calibration_show_reference_box(
        image: Image,
        calibration_show_reference_box: bool = True,
        calibration_reference_box_color="red",
        **kwargs
    ):
        if (
            calibration_show_reference_box
            and isinstance(image, Image)
            and image.has_meta("calibration_reference_box")
        ):
            plt.plot(
                *image.get_meta("calibration_reference_box").exterior.xy,
                # linewidth=multimodule_boxes_linewidth,
                color=calibration_reference_box_color,
            )

    register_show_plugin(calibration_show_reference_box)

    def segment_module_show_box(
        image: Image,
        segment_module_show_box: bool = True,
        segment_module_show_box_color="red",
        **kwargs
    ):
        if (
            segment_module_show_box
            and isinstance(image, Image)
            and image.has_meta("segment_module_original_box")
        ):
            plt.plot(
                *image.get_meta("segment_module_original_box").exterior.xy,
                color=segment_module_show_box_color,
            )

    register_show_plugin(segment_module_show_box)

    def show_image(
        image: Image,
        clip_low: float = 0.001,
        clip_high: float = 99.999,
        colorbar: bool = True,
        **kwargs
    ):
        clip_low = clip_low if clip_low is not None else 0.0
        clip_high = clip_high if clip_high is not None else 100.0
        p = np.percentile(image._data, [clip_low, clip_high])
        d = np.clip(image._data, p[0], p[1])
        plt.imshow(d, cmap="gray")
        if colorbar:
            plt.colorbar()

    register_show_plugin(show_image, -100)

    def axis_options(
        image: Image,
        show_axis: bool = True,
        show_title: bool = True,
        max_title_length: bool = 30,
        **kwargs
    ):
        if not show_axis:
            plt.axis("off")
        if show_title:
            if isinstance(image, CellImage):
                t = "{} (r: {:d}, c: {:d})".format(
                    str(image.path.name), image.row, image.col
                )
            elif image.path is not None:
                t = str(image.path.name)
            else:
                t = ""

            if len(t) > max_title_length:
                l1 = max_title_length // 2 - 2
                l2 = max_title_length - l1 - 2
                t = t[:l1] + ".." + t[len(t) - l2 :]

            plt.title(t)

    register_show_plugin(axis_options, -200)


class _Base:

    T = TypeVar("T")

    @classmethod
    def from_other(
        cls: Type[T], other: T, drop_meta_types: List[Type] = None, **kwargs
    ) -> T:
        """Create a new image by partially overwriting the properties of another

        Args:
            other (Image): The other image
            drop_meta_types (List[Type]): Drop any meta attributes that are insteanceof these types
            **kwargs: Arguments that should be overwritten
        """
        required = inspect.getfullargspec(cls.__init__)[0]

        if "meta" in kwargs.keys() and isinstance(kwargs["meta"], dict):
            kwargs["meta"] = pd.Series(kwargs["meta"])

        other_args = dict()
        for name in required:
            if name == "meta" and "meta" in kwargs.keys():
                # joint meta dictionaries
                tmp = copy.copy(other._meta)
                if drop_meta_types is not None:
                    tmp = pd.Series(
                        {
                            k: v
                            for k, v in tmp.items()
                            if not np.any([isinstance(v, x) for x in drop_meta_types])
                        }
                    )
                kwargs["meta"] = kwargs["meta"].combine_first(tmp)

            if name not in kwargs.keys() and name != "self":

                # first, try public property, then private property, then meta attribute
                if name == "data":
                    other_args[name] = other._data
                elif hasattr(other, name):
                    other_args[name] = getattr(other, name)
                elif hasattr(other, "_" + name):
                    other_args[name] = getattr(other, "_" + name)
                elif isinstance(other, Image) and other.has_meta(name):
                    other_args[name] = other.get_meta(name)

                if name == "meta" and drop_meta_types is not None:
                    other_args[name] = {
                        k: v
                        for k, v in other_args[name].items()
                        if not np.any([isinstance(v, x) for x in drop_meta_types])
                    }

        return cls(**kwargs, **other_args)

    def from_self(self: T, drop_meta_types: List[Type] = None, **kwargs) -> T:
        return type(self).from_other(self, drop_meta_types=drop_meta_types, **kwargs)


class Image(_Base):
    """A general image"""

    @staticmethod
    def _map_numpy_dtype(dtype):
        if dtype == np.float32 or dtype == np.float64:
            return DType.FLOAT
        elif dtype == np.uint16 or dtype == np.uint32 or dtype == np.uint64:
            return DType.UNSIGNED_INT
        elif dtype == np.uint8:
            return DType.UNSIGNED_BYTE
        elif (
            dtype == np.int8
            or dtype == np.int16
            or dtype == np.int32
            or dtype == np.int64
        ):
            return DType.INT

    @staticmethod
    def _unify_dtypes(array):
        if (
            Image._map_numpy_dtype(array.dtype) == DType.UNSIGNED_INT
            and array.dtype != DTYPE_UNSIGNED_INT
        ):
            if (
                array.max() > np.iinfo(DTYPE_UNSIGNED_INT).max
                or array.min() < np.iinfo(DTYPE_UNSIGNED_INT).min
            ):
                raise RuntimeError(
                    "Datatype conversion to {} failed, since original data exceeds dtype limits.".format(
                        DTYPE_UNSIGNED_INT
                    )
                )
            return array.astype(DTYPE_UNSIGNED_INT)
        if (
            Image._map_numpy_dtype(array.dtype) == DType.INT
            and array.dtype != DTYPE_INT
        ):
            if (
                array.max() > np.iinfo(DTYPE_INT).max
                or array.min() < np.iinfo(DTYPE_INT).min
            ):
                raise RuntimeError(
                    "Datatype conversion to {} failed, since original data exceeds dtype limits.".format(
                        DTYPE_INT
                    )
                )
            return array.astype(DTYPE_INT)
        if (
            Image._map_numpy_dtype(array.dtype) == DType.FLOAT
            and array.dtype != DTYPE_FLOAT
        ):
            return array.astype(DTYPE_FLOAT)

        # default
        return array

    class LazyData:
        @classmethod
        @lru_cache(maxsize=SEQUENCE_MAX_CACHE_SIZE, typed=True)
        def _load(
            cls,
            load_fn: Callable[[], np.ndarray],
            checks: Tuple[Callable[[np.ndarray], np.ndarray]],
        ) -> np.ndarray:
            data = load_fn()

            # perform data checks/conversions
            for check in checks:
                data = check(data)

            # make it immutable
            data.setflags(write=False)

            return data

        def __init__(self, load_fn: Callable[[], np.ndarray]):
            self._load_fn = load_fn
            self._checks: List[Callable[[np.ndarray], np.ndarray]] = list()

        def __getattr__(self, name: str):
            # forward to numpy
            data = self._load(self._load_fn, tuple(self._checks))
            return getattr(data, name)

        def __getitem__(self, s):
            data = self._load(self._load_fn, tuple(self._checks))
            return data[s]

        def push_check(self, fn: Callable[[np.ndarray], np.ndarray]):
            self._checks.append(fn)

        def load(self) -> np.ndarray:
            return self._load(self._load_fn, tuple(self._checks))

    def __init__(
        self,
        data: np.ndarray,
        path: Path = None,
        modality: Modality = None,
        meta: Union[Dict[str, Any], pd.Series] = None,
    ):
        """Create a new image. All non-float images as automatically converted to uint.

        Args:
            data (np.ndarray): The image data
            path (Path): Path to the image
            modality (Modality): The imaging modality
            meta (Dict[str, Any]): Meta attributes of this image
        """
        self._data = data
        self._meta = (
            meta
            if isinstance(meta, pd.Series)
            else pd.Series(meta)
            if meta is not None
            else pd.Series()
        )
        self._meta["modality"] = modality
        self._meta["path"] = path.absolute() if path is not None else None

        if isinstance(data, np.ndarray):
            self._data = Image._unify_dtypes(self._data)
            self._data.setflags(write=False)
        else:
            self._data.push_check(Image._unify_dtypes)

    def show(self, **kwargs):
        """Show this image"""
        _invoke_show_plugins(self, **kwargs)

    _T = TypeVar("T")

    def as_type(self: _T, dtype: DType) -> _T:
        if dtype == DType.FLOAT:
            return self.from_self(data=img_as_float(self._data))
        elif dtype == DType.UNSIGNED_INT:
            return self.from_self(data=img_as_uint(self._data))
        elif dtype == DType.INT:
            return self.from_self(data=img_as_int(self._data))

    def __add__(self: _T, other: _T) -> _T:
        if self.dtype != other.dtype:
            raise RuntimeError("Images must have the same datatype")
        return self.from_self(data=self._data + other._data)

    def __sub__(self: _T, other: _T) -> _T:
        if self.dtype != other.dtype:
            raise RuntimeError("Images must have the same datatype")
        if self.dtype == DType.UNSIGNED_INT:
            res = self._data.astype(DTYPE_INT) - other._data.astype(DTYPE_INT)
            iinfo = np.iinfo(DTYPE_UNSIGNED_INT)
            res = np.clip(res, 0, iinfo.max)
            return self.from_self(data=res)
        else:
            return self.from_self(data=self._data - other._data)

    def __mul__(self: _T, other: _T) -> _T:
        if self.dtype != other.dtype:
            raise RuntimeError("Images must have the same datatype")
        return self.from_self(data=self._data * other._data)

    def __truediv__(self: _T, other: _T) -> _T:
        if self.dtype != DType.FLOAT or other.dtype != DType.FLOAT:
            raise RuntimeError("Images must be of type float")
        return self.from_self(data=self._data / other._data)

    def __floordiv__(self: _T, other: _T) -> _T:
        if self.dtype != other.dtype:
            raise RuntimeError("Images must have the same datatype")
        return self.from_self(data=self._data // other._data)

    def __mod__(self: _T, other: _T) -> _T:
        if self.dtype != other.dtype:
            raise RuntimeError("Images must have the same datatype")
        return self.from_self(data=self._data % other._data)

    def __pow__(self: _T, other: _T) -> _T:
        if self.dtype != other.dtype:
            raise RuntimeError("Images must have the same datatype")
        return self.from_self(data=self._data ** other._data)

    def __deepcopy__(self: _T, memo) -> _T:
        # let behavior be determined by overridden attributes
        return type(self).from_other(self)

    @property
    def data(self) -> np.ndarray:
        """The underlying image data"""
        if isinstance(self._data, Image.LazyData):
            v = self._data.load()
        else:
            v = self._data.view()
        return v

    @property
    def path(self) -> Path:
        """Path to the original image"""
        return self.get_meta("path")

    @property
    def dtype(self) -> DType:
        """Datatype of the image"""
        return Image._map_numpy_dtype(self._data.dtype)

    @property
    def shape(self) -> Tuple[int, int]:
        """Shape of the image"""
        return copy.deepcopy(self.data.shape)

    @property
    def modality(self) -> Modality:
        """The imaging modality"""
        return self.get_meta("modality")

    @property
    def lazy(self) -> bool:
        """Check, if this is lazy loaded"""
        return isinstance(self._data, Image.LazyData)

    def get_meta(self, key: str) -> Any:
        """Access a meta attribute"""
        if isinstance(self._meta[key], np.ndarray):
            v = self._meta[key].view()
            v.setflags(write=False)
            return v
        else:
            return copy.copy(self._meta[key])

    def has_meta(self, key: str) -> bool:
        """Check if a meta attribute is set"""
        return key in self._meta.index

    def list_meta(self) -> List[str]:
        """List avaliable meta keys"""
        return list(self._meta.index)

    def meta_from_path(
        self,
        pattern: str,
        key: str,
        target_type: Type,
        group_n: int = 0,
        transform: Callable[[Any], Any] = None,
    ) -> Image:
        """Extract meta information from path. The group_n'th matching group
        from pattern is used as meta value

        Args:
            pattern (str): Regular expression used to parse meta
            key (str): Key of the meta attribute
            target_type (Type): Result is converted to this datatype
            group_n (int): Index of matching group
            transform (Callable[[Any], Any]): Optional function that is applied on the value
                before datatype conversion

        Returns:
            image (Image): Resulting Image
        """
        s = str(self.path.absolute())
        res = re.search(pattern, s)
        v = res.group(group_n)
        if transform is not None:
            v = transform(v)
        v = target_type(v)
        return self.from_self(meta={key: v})

    def meta_from_fn(self, fn: Callable[[Image], Dict[str, Any]]) -> Image:
        """Extract meta data using given callable

        Args:
            fn (Callable[[Image], Dict[str, Any]]): Function used to extract meta data
        
        Returns
            image (Image): Resulting Image
        """
        return self.from_other(self, meta=fn(self))

    def _meta_to_pandas(self) -> pd.Series:
        """Convert (compatible) meta data to pandas series"""
        return self._meta

    def meta_to_pandas(self) -> pd.Series:
        """Convert (compatible) meta data to pandas series"""
        return copy.deepcopy(self._meta_to_pandas())


class ImageSequence(_Base):
    """An immutable sequence of images, allowing for access to single images as well as analysis of the sequence"""

    def _show(self, imgs: List[Image], cols: int, *args, **kwargs):
        N = len(imgs)
        rows = math.ceil(N / cols)

        # adjust the figure size
        if self.shape is not None:
            aspect = self.shape[0] / self.shape[1]
        else:
            aspect = 1.0
        plt.figure(figsize=(6 * cols, 6 * rows * aspect))

        for i, img in enumerate(imgs):
            plt.subplot(rows, cols, i + 1)
            img.show(*args, **kwargs)

    class _PandasHandler:
        def __init__(self, parent: ImageSequence):
            self._parent = parent
            pass

        class _Sub:
            def _result(self, pandas_result):
                if isinstance(pandas_result, pd.DataFrame):
                    idx = pandas_result.index.to_list()
                    result = [self._parent._images[i] for i in idx]
                    seq = type(self._parent).from_other(self._parent, images=result)
                    seq._meta_df = pandas_result.reset_index(drop=True)
                    return seq
                elif isinstance(pandas_result, pd.Series):
                    idx = pandas_result.name
                    return self._parent[idx]

            def __init__(self, parent: ImageSequence, attr):
                self._parent = parent
                self._attr = attr

            def __call__(self, *argv, **kwargs):
                pandas_result = self._attr(*argv, **kwargs)
                return self._result(pandas_result)

            def __getitem__(self, arg):
                pandas_result = self._attr[arg]
                return self._result(pandas_result)

        def __getattr__(self, name):
            attr = getattr(self._parent.meta_to_pandas(), name)
            return self._Sub(self._parent, attr)

    def __init__(
        self,
        images: List[Image],
        same_camera: bool = False,
        allow_different_dtypes=False,
    ):
        """Initialize a module image sequence
        
        Args:
            images (List[Image]): The list of images
            came_camera (bool): Indicates, if all images are from the same camera and hence share the same intrinsic parameters
            allow_different_dtypes (bool): Allow images to have different datatypes?
        """

        self._images = images
        self._same_camera = same_camera
        self._allow_different_dtypes = allow_different_dtypes
        self._meta_df = None
        if len(self.images) == 0:
            logging.error("Creation of an empty sequence is not supported")
            exit()

        # check that all have the same modality, dimension, dtype and module configuration
        shape = self.images[0].shape
        dtype = self.images[0].dtype
        modality = self.images[0].modality
        # for img in self.images:
        #    if img.dtype != dtype and not allow_different_dtypes:
        #        logging.error(
        #            'Cannot create sequence from mixed dtypes. Consider using the "allow_different_dtypes" argument, when reading images.'
        #        )
        #        exit()
        #    if img.shape != shape and same_camera:
        #        logging.error(
        #            'Cannot create sequence from mixed shapes. Consider using the "same_camera" argument, when reading images.'
        #        )
        #        exit()
        #    if img.modality != modality:
        #        logging.error("Cannot create a sequence from mixed modalities.")
        #        exit()

        # namespace for accessing pandas methods
        self.pandas = self._PandasHandler(self)

    def head(self, N: int = 4, cols: int = 2, *args, **kwargs):
        """Show the first N images

        Args:
            N (int): Number of images to show
            cols (int): How many images to show in a column
        """
        self._show(self.images[:N], cols, *args, **kwargs)

    def tail(self, N: int = 4, cols: int = 2, *args, **kwargs):
        """Show the last N images

        Args:
            N (int): Number of images to show
            cols (int): How many images to show in a column
        """
        self._show(self.images[-N:], cols, *args, **kwargs)

    _T = TypeVar("T")

    def apply(
        self, fn: Callable[[Image], Image], *argv, progress_bar: bool = True, **kwargs
    ) -> ImageSequence:
        """Apply the given callable on every image. Returns a copy of the
        original sequence

        Args:
            fn (Callable[[Image], Image]): Callable that receives and returns an Image
            progress_bar (bool): Show progress bar?
        
        Returns:
            sequence (ImageSequence): The copy with modified images
        """
        result = []
        p = tqdm if progress_bar else lambda x: x
        for img in p(self._images):
            result.append(fn(img, *argv, **kwargs))
        return self.from_self(images=result)

    def apply_image_data(
        self: _T,
        fn: Callable[[np.ndarray], np.ndarray],
        *argv,
        progress_bar: bool = True,
        **kwargs
    ) -> _T:
        """Apply the given callable on every image data. Returns a copy of the
        original sequence with modified data

        Args:
            fn (Callable[[np.ndarray], np.ndarray]): Callable that receives a np.ndarray
                and returns a np.ndarray. Note that the argument is immutable.
            progress_bar (bool): Show progress bar?
        
        Returns:
            sequence (ImageSequence): The copy with modified data
        """
        result = []
        p = tqdm if progress_bar else lambda x: x
        for img in p(self._images):
            data = img.data
            res = fn(data, *argv, **kwargs)
            result.append(type(img).from_other(img, data=res))
        return self.from_self(images=result)

    def meta_from_path(
        self,
        pattern: str,
        key: str,
        target_type: Type,
        group_n: int = 1,
        transform: Callable[[Any], Any] = None,
    ) -> ImageSequence:
        """Extract meta information from path of individual aimges. The group_n'th matching group
        from pattern is used as meta value

        Args:
            pattern (str): Regular expression used to parse meta
            key (str): Key of the meta attribute
            target_type (Type): Result is converted to this datatype
            group_n (int): Index of matching group
            transform (Callable[[Any], Any]): Optional function that is applied on the value
                before datatype conversion

        Returns:
            images (ImageSequence): Resulting ImageSequence
        """
        result = []
        for img in self._images:
            result.append(
                img.meta_from_path(
                    pattern=pattern,
                    key=key,
                    target_type=target_type,
                    group_n=group_n,
                    transform=transform,
                )
            )
        return self.from_self(images=result)

    def meta_from_fn(
        self, fn: Callable[[Image], Dict[str, Any]], **kwargs
    ) -> ImageSequence:
        """Extract meta information using given function

        Args:
            fn (Callable[[Image], Dict[str, Any]]): Function that is applied on every element of the sequence
        """
        return self.apply(fn=lambda x: x.meta_from_fn(fn), **kwargs)

    def meta_to_pandas(self) -> pd.DataFrame:
        """Convert meta from images to pandas DataFrame"""
        if self._meta_df is None:
            series = [img._meta_to_pandas() for img in self._images]
            self._meta_df = pd.DataFrame(data=series)
            self._meta_df = self._meta_df.astype(
                {"modality": str}
            )  # allow for easy comparison
        return self._meta_df.copy()  # pd.DataFrame has no writable flag :(

    def as_type(self: _T, dtype: DType) -> _T:
        """Convert sequence to specified dtype"""
        result = [img.as_type(dtype) for img in self._images]
        return self.from_self(images=result)

    def __add__(self: _T, other: _T) -> _T:
        res = [x + y for x, y in zip(self.images, other.images)]
        return self.from_self(images=res)

    def __sub__(self: _T, other: _T) -> _T:
        res = [x - y for x, y in zip(self.images, other.images)]
        return self.from_self(images=res)

    def __mul__(self: _T, other: _T) -> _T:
        res = [x * y for x, y in zip(self.images, other.images)]
        return self.from_self(images=res)

    def __truediv__(self: _T, other: _T) -> _T:
        res = [x / y for x, y in zip(self.images, other.images)]
        return self.from_self(images=res)

    def __floordiv__(self: _T, other: _T) -> _T:
        res = [x // y for x, y in zip(self.images, other.images)]
        return self.from_self(images=res)

    def __mod__(self: _T, other: _T) -> _T:
        res = [x % y for x, y in zip(self.images, other.images)]
        return self.from_self(images=res)

    def __pow__(self: _T, other: _T) -> _T:
        res = [x ** y for x, y in zip(self.images, other.images)]
        return self.from_self(images=res)

    @property
    def images(self) -> List[Image]:
        """Access the list of images"""
        return self._images

    @property
    def dtype(self) -> DType:
        """Access the image datatype"""
        return self.images[0].dtype if not self._allow_different_dtypes else None

    @property
    def shape(self) -> Tuple[int, int]:
        """Access the image shape"""
        return self.images[0].shape if self._same_camera else None

    @property
    def modality(self) -> Modality:
        """Access the imaging modaility"""
        return self.images[0].modality

    @property
    def same_camera(self) -> bool:
        """Indicate, if the images originate from the same camera"""
        return self._same_camera

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, i: int) -> Image:
        return self.images[i]


ImageOrSequence = Union[Image, ImageSequence]


class CellImage(Image):
    """An image of a solar cell with additional meta data"""

    def __init__(
        self,
        data: np.ndarray,
        modality: Modality,
        row: int,
        col: int,
        path: Path = None,
        meta: Dict[str, Any] = None,
    ):
        """Initialize a cell image

        Args:
            data (np.ndarray): The image data
            modality (Modality): The imaging modality
            path (Path): Path to the image
            row (int): Row index (zero-based)
            col (int): Cell index (zero-based)
            meta (Dict[str, Any]): Meta data
        """

        super().__init__(data, path=path, modality=modality, meta=meta)
        self._meta["row"] = row
        self._meta["col"] = col

    @property
    def row(self) -> int:
        """0-based row index of the cell in the original module"""
        return self.get_meta("row")

    @property
    def col(self) -> int:
        """0-based column index of the cell in the original module"""
        return self.get_meta("col")

    @property
    def path(self) -> Path:
        """Get this images path"""
        p = super().path
        return Path(
            "{}_row{:02d}_col{:02d}{}".format(p.stem, self.row, self.col, p.suffix)
        )

    def show(self, *argv, **kwargs):
        """Show this image"""
        super().show(*argv, **kwargs)


class CellImageSequence(ImageSequence):
    """An immutable sequence of cell images, allowing for access to single images as well as analysis of the sequence"""

    def __init__(self, images: List[CellImage]):
        """Initialize a module image sequence
        
        Args:
            images (List[CellImage]): The list of images
        """

        super().__init__(images, False)


class ModuleImage(Image):
    """An image of a solar module with additional meta data"""

    def __init__(
        self,
        data: np.ndarray,
        modality: Modality,
        path: Path = None,
        cols: int = None,
        rows: int = None,
        meta: dict = None,
    ):
        """Initialize a module image

        Args:
            data (np.ndarray): The image data
            modality (Modality): The imaging modality
            path (Path): Path to the image
            cols (int): Number of cells in a column
            rows (int): Number of cells in a row
        """

        super().__init__(data, path, modality, meta)
        self._meta["cols"] = cols
        self._meta["rows"] = rows

    def grid(self) -> np.ndarray:
        """Create a grid of corners according to the module geometry
        
        Returns:
            grid: (cols*rows, 2)-array of coordinates on a regular grid
        """

        if self.cols is not None and self.rows is not None:
            x, y = np.mgrid[0 : self.cols + 1 : 1, 0 : self.rows + 1 : 1]
            grid = np.stack([x.flatten(), y.flatten()], axis=1)
            return grid
        else:
            logging.error("Module geometry is not initialized")
            exit()

    @property
    def cols(self):
        """Number of cell-columns"""
        return self.get_meta("cols")

    @property
    def rows(self):
        """Number of row-columns"""
        return self.get_meta("rows")


class PartialModuleImage(ModuleImage):
    """An image of a solar module with additional meta data"""

    def __init__(
        self,
        data: np.ndarray,
        modality: Modality,
        path: Path = None,
        cols: int = None,
        rows: int = None,
        first_col: int = None,
        first_row: int = None,
        meta: dict = None,
    ):
        """Initialize a module image

        Args:
            data (np.ndarray): The image data
            modality (Modality): The imaging modality
            path (Path): Path to the image
            cols (int): Number of completely visible cells in a column
            rows (int): Number of completely visible cells in a row
            first_col (int): Index of the first complete column shown
            first_row (int): Index of the first complete row shown
        """

        super().__init__(data, modality, path, cols, rows, meta)

        self._meta["first_col"] = first_col
        self._meta["first_row"] = first_row

    def grid(self) -> np.ndarray:
        """Create a grid of corners according to the module geometry
        
        Returns:
            grid: (cols*rows, 2)-array of coordinates on a regular grid
        """

        if self.cols is not None and self.rows is not None:
            x, y = np.mgrid[0 : self.cols + 1 : 1, 0 : self.rows + 1 : 1]
            if self.first_col and self.first_row:
                x += self.first_col
                y += self.first_row
            grid = np.stack([x.flatten(), y.flatten()], axis=1)
            return grid
        else:
            logging.error("Module geometry is not initialized")
            exit()

    @property
    def first_col(self) -> Optional[int]:
        if self.has_meta("first_col"):
            return self.get_meta("first_col")
        else:
            return None

    @property
    def first_row(self) -> Optional[int]:
        if self.has_meta("first_row"):
            return self.get_meta("first_row")
        else:
            return None


ModuleOrPartialModuleImage = Union[ModuleImage, PartialModuleImage]


class ModuleImageSequence(ImageSequence):
    """An immutable sequence of module images, allowing for access to single images as well as analysis of the sequence"""

    def __init__(
        self,
        images: List[ModuleOrPartialModuleImage],
        same_camera: bool = False,
        allow_different_dtypes=False,
    ):
        """Initialize a module image sequence
        
        Args:
            images (List[ModuleImage]): The list of images
            same_camera (bool): Indicates if all images are from the same camera
            allow_different_dtypes (bool): Allow images to have different datatypes?
        """

        cols = images[0].cols
        rows = images[0].rows
        for img in images:
            if img.cols != cols:
                logging.error(
                    "Cannot create sequence from different module configurations"
                )
                exit()
            if img.rows != rows:
                logging.error(
                    "Cannot create sequence from different module configurations"
                )
                exit()

        super().__init__(images, same_camera, allow_different_dtypes)


ModuleImageOrSequence = Union[
    ModuleImageSequence, ModuleImage, PartialModuleImage, Image
]


def _sequence(*args):
    """Assure that the first argument is a sequence and handle the first return value accordingly"""

    def decorator_sequence(func):
        @wraps(func)
        def wrapper_sequence(*args, **kwargs):
            if not isinstance(args[0], ImageSequence):
                args = list(args)
                args[0] = (
                    ModuleImageSequence([args[0]], same_camera=False)
                    if type(args[0]) == ModuleImage
                    else ImageSequence([args[0]], same_camera=False)
                )
                unwrap = True
            else:
                unwrap = False
            res = func(*tuple(args), **kwargs)
            if unwrap and not disable_unwrap:
                if isinstance(res, tuple) and isinstance(res[0], ImageSequence):
                    res[0] = res[0].images[0]
                elif isinstance(res, ImageSequence):
                    res = res.images[0]
            return res

        return wrapper_sequence

    if len(args) == 1 and callable(args[0]):
        disable_unwrap = False
        return decorator_sequence(args[0])
    else:
        disable_unwrap = args[0] if len(args) == 1 else False
        return decorator_sequence



"""Provides access to demo datasets"""

from .image import (
    ModuleImageSequence,
    ModuleImage,
    EL_IMAGE,
    ImageSequence,
    CellImage,
    CellImageSequence,
)
from .io import *
from pathlib import Path
from google_drive_downloader import GoogleDriveDownloader as gdd
from typing import Tuple, Dict
import os
import requests
from zipfile import ZipFile
import logging

_DS_PATH = Path(__file__).parent.absolute() / "datasets"
_DS_KEYS = {
    "20191219_poly10x6": "1B5fQPLvStuMvuYJ5CxbzyxfwuWQdfNVE",
    "20200728_elpv_labels": "1hK_hViiZ1-rHhvI3yGxpC6DSCXAyAFiJ",
}
_ZIP_DS_URLS = {"elpv": "https://github.com/zae-bayern/elpv-dataset/archive/master.zip"}


def _get_dataset_key(name: str):
    if name in _DS_KEYS.keys():
        return _DS_KEYS[name]
    else:
        keys = os.getenv("PVINSPECT_KEYS").split(";")
        keys = {x.split(",")[0]: x.split(",")[1] for x in keys}
        if name in keys.keys():
            return keys[name]
        else:
            raise RuntimeError(
                'The specified dataset "{}" could not be found. Maybe you tried \
                to access a protected dataset and didn\'t set PVINSPECT_KEYS environment variable?'
            )


def _check_and_download_ds(name: str):
    ds_path = Path(__file__).parent.absolute() / "datasets" / name
    if not ds_path.is_dir():
        logging.info("Data is being downloaded..")
        k = _get_dataset_key(name)
        ds_path.mkdir(parents=True, exist_ok=False)
        gdd.download_file_from_google_drive(k, str(ds_path / "data.zip"), unzip=True)
    return ds_path


def _check_and_download_zip_ds(name: str) -> Path:
    url = _ZIP_DS_URLS[name]
    target = _DS_PATH / name

    if not target.is_dir():
        logging.info("Data is being downloaded..")
        target.mkdir()
        r = requests.get(url, allow_redirects=True)
        open(target / "data.zip", "wb").write(r.content)
        zipf = ZipFile(target / "data.zip")
        zipf.extractall(target)

    return target


def poly10x6(N: int = 0) -> ModuleImageSequence:
    """Read sequence of 10x6 poly modules
    
    Args:
        N (int): Only read first N images
    """
    p = _check_and_download_ds("20191219_poly10x6")
    return read_module_images(p, EL_IMAGE, True, 10, 6, N=N)


def elpv(N: int = 0) -> ImageSequence:
    """Read images from ELPV dataset

        Note:
            This dataset is part of the following publication:
            Deitsch, Sergiu, et al. "Automatic classification of defective photovoltaic module cells in electroluminescence images."
            Solar Energy, Elsevier BV, 2019, 185, 455-468. 
            
            Additional labels for defect types are provided by the author of this toolbox.
        
        Args:
            N (int): Number of images to return. Defaults to using all images.

        Returns:
            images: Images from the ELPV dataset with defect type annotations as `Image` meta data
    """
    # download and read images
    images_path = _check_and_download_zip_ds("elpv") / "elpv-dataset-master" / "images"
    seq = read_images(images_path, same_camera=False, modality=EL_IMAGE, N=N)

    # download and read labels
    labels_path = _check_and_download_ds("20200728_elpv_labels") / "labels.csv"

    logging.info("Loading labels..")
    labels = pd.read_csv(
        labels_path,
        delimiter=";",
        index_col="filename",
        dtype={
            "defect probability": float,
            "wafer": str,
            "crack": bool,
            "inactive": bool,
            "blob": bool,
            "finger": bool,
            "testset": bool,
        },
    ).rename(columns={"defect probability": "defect_probability"})

    # associate images with labels
    def label(img: Image):
        l = labels.loc["images/{}".format(img.path.name)]
        return l.to_dict()

    # read images and labels
    seq = seq.meta_from_fn(label, progress_bar=False)

    return seq


def caip_dataB() -> Tuple[ModuleImageSequence, ModuleImageSequence, ObjectAnnotations]:
    """Read DataB from CAIP paper (private dataset)
    
    Note:
        This dataset is from the following publication:
        Hoffmann, Mathis, et al. "Fast and robust detection of solar modules in electroluminescence images."
        International Conference on Computer Analysis of Images and Patterns. Springer, Cham, 2019.

    Returns:
        images1: All modules with shape 10x6
        images2: All modules with shape 9x4
        annot: Annotations specifying the position of modules
    """
    p = _check_and_download_ds("20200616_caip_v2")
    images1 = read_module_images(
        p / "deitsch_testset" / "10x6",
        EL_IMAGE,
        False,
        10,
        6,
        allow_different_dtypes=True,
    )
    images2 = read_module_images(
        p / "deitsch_testset" / "9x4",
        EL_IMAGE,
        False,
        9,
        4,
        allow_different_dtypes=True,
    )
    annot = load_json_object_masks(p / "deitsch_testset" / "module_locations.json")
    return images1, images2, annot


def caip_dataC() -> Tuple[ModuleImageSequence, ObjectAnnotations]:
    """Read DataC from CAIP paper (private dataset)
    
    Note:
        This dataset is from the following publication:
        Hoffmann, Mathis, et al. "Fast and robust detection of solar modules in electroluminescence images."
        International Conference on Computer Analysis of Images and Patterns. Springer, Cham, 2019.

    Returns:
        images: All modules images
        annot: Annotations specifying the position of modules
    """
    p = _check_and_download_ds("20200616_caip_v2")
    annot = load_json_object_masks(p / "multiple" / "module_locations.json")
    return read_module_images(p / "multiple", EL_IMAGE, True, 10, 6), annot


def caip_dataD() -> Tuple[ModuleImageSequence, ObjectAnnotations]:
    """Read DataC from CAIP paper (private dataset)
    
    Note:
        This dataset is from the following publication:
        Hoffmann, Mathis, et al. "Fast and robust detection of solar modules in electroluminescence images."
        International Conference on Computer Analysis of Images and Patterns. Springer, Cham, 2019.

    Returns:
        images: All modules images
        annot: Annotations specifying the position of modules
    """
    p = _check_and_download_ds("20200616_caip_v2")
    annot = load_json_object_masks(p / "rotated" / "module_locations.json")
    return read_module_images(p / "rotated", EL_IMAGE, True, 10, 6), annot


def stitching_demo(N: int = 0) -> List[Tuple[Image, Image]]:
    """Data to demonstrate stitching capabilities

    Args:
        N (int): Number of image pairs that is returened

    Returns:
        images: List of image pairs
    """
    result = list()
    images = poly10x6(N)

    for image in images:
        height = image.shape[0]

        # split
        img0 = Image(image.data[0 : int(height * 2 // 3)])
        img1 = Image(image.data[int(height // 3) :])
        result.append((img0, img1))

    return result


def calibration_ipv40CCD_FF(N: int = 0) -> Dict[str, ImageSequence]:
    """Flat-field calibration data for ipv40CCD (private dataset)

    Args:
        N (int): Number of images per excitation

    Returns:
        images: Dict with excitation as key and images
    """
    p = _check_and_download_ds("20200303_calibration_iPV40CCD")
    res = dict()
    for d in p.glob("FF*"):
        key = d.name.split("_")[1]
        seq = read_images(path=d, same_camera=False, N=N)
        res[key] = seq
    return res


def calibration_ipv40CCD_distortion(N: int = 0) -> ImageSequence:
    """Lens calibration data for ipv40CCD (private dataset)

    Args:
        N (int): Number of images

    Returns:
        images: Sequence of images
    """
    p = _check_and_download_ds("20200303_calibration_iPV40CCD")
    return read_images(path=p / "distortion", same_camera=True, N=N)


def multi_module_detection(N: int = 0) -> Tuple[ObjectAnnotations, ImageSequence]:
    """Dataset for multi module detection (private dataset)

    Args:
        N (int): Number of images

    Returns:
        anns: Dict of annotations by image
        imgs: Sequence of images
    """
    p = _check_and_download_ds("20200331_multi_module_detection")
    imgs = read_images(
        path=p, same_camera=False, N=N, pattern="**/*.png", modality=EL_IMAGE
    )
    anns = load_json_object_masks(path=p / "labels.json")
    return anns, imgs



"""Data loading and viewing"""

from . import datasets
from . import io
from . import image
from .image import (
    EL_IMAGE,
    PL_IMAGE,
    DType,
    _register_default_plugins,
    Image,
    ImageSequence,
    ModuleImage,
    ModuleImageSequence,
    PartialModuleImage,
    CellImage,
    CellImageSequence,
)

__all__ = ["EL_IMAGE", "PL_IMAGE", "DType"]

_register_default_plugins()



from pvinspect import data
from pvinspect.preproc import detection
from pvinspect.data.image import *
from pathlib import Path
from pvinspect.common.transform import HomographyTransform, FullTransform
from pvinspect.common.util import objdetect_metrics
import numpy as np
from test.utilities import assert_equal


def test_locate_homography():
    seq = data.datasets.poly10x6(2)
    seq = detection.locate_module_and_cells(seq, False)

    assert isinstance(seq[0].get_meta("transform"), HomographyTransform)
    assert isinstance(seq[1].get_meta("transform"), HomographyTransform)
    assert seq[0].get_meta("transform").valid
    assert seq[1].get_meta("transform").valid

    # check correct origin
    x = seq[0].get_meta("transform")(np.array([[0.0, 0.0]])).flatten()
    assert x[0] > 1760 and x[0] < 1840
    assert x[1] > 80 and x[1] < 160
    x = seq[1].get_meta("transform")(np.array([[0.0, 0.0]])).flatten()
    assert x[0] > 1760 and x[0] < 1840
    assert x[1] > 80 and x[1] < 160


def test_locate_full():
    seq = data.datasets.poly10x6(2)
    seq = detection.locate_module_and_cells(seq, True)

    assert isinstance(seq[0].get_meta("transform"), FullTransform)
    assert isinstance(seq[1].get_meta("transform"), FullTransform)
    assert seq[0].get_meta("transform").valid
    assert seq[1].get_meta("transform").valid

    # check correct origin
    x = seq[0].get_meta("transform")(np.array([[0.0, 0.0]])).flatten()
    assert x[0] > 1760 and x[0] < 1840
    assert x[1] > 80 and x[1] < 160
    x = seq[1].get_meta("transform")(np.array([[0.0, 0.0]])).flatten()
    assert x[0] > 1760 and x[0] < 1840
    assert x[1] > 80 and x[1] < 160


def test_segment_cells():
    seq = data.datasets.poly10x6(2)
    seq = detection.locate_module_and_cells(seq, True)
    cells = detection.segment_cells(seq)

    assert isinstance(cells, CellImageSequence)
    assert len(cells) == 120
    assert isinstance(cells[0], CellImage)
    assert cells[0].row == 0
    assert cells[0].col == 0
    assert cells[1].col == 1
    assert cells[11].row == 1
    assert cells[0].has_meta("segment_module_original")
    assert (
        cells[0]
        .get_meta("segment_module_original")
        .has_meta("segment_module_original_box")
    )


def test_segment_cells_single_image():
    seq = data.datasets.poly10x6(1)
    seq = detection.locate_module_and_cells(seq, True)
    cells = detection.segment_cells(seq[0])

    assert isinstance(cells, CellImageSequence)
    assert len(cells) == 60
    assert isinstance(cells[0], CellImage)


def test_segment_modules():
    seq = data.datasets.poly10x6(2)
    seq = detection.locate_module_and_cells(seq, True)
    modules = detection.segment_modules(seq)

    assert isinstance(modules, ModuleImageSequence)
    assert isinstance(modules[0], ModuleImage)
    assert len(modules) == 2
    assert modules[0].path == seq[0].path
    assert modules.same_camera is False

    x = modules[0].get_meta("transform")(np.array([[0.0, 0.0]])).flatten()
    assert_equal(x[0], 0.0)
    assert_equal(x[1], 0.0)
    x = modules[0].get_meta("transform")(np.array([[10.0, 0.0]])).flatten()
    assert_equal(x[0], modules[0].shape[1])
    assert_equal(x[1], 0.0)
    x = modules[0].get_meta("transform")(np.array([[10.0, 6.0]])).flatten()
    assert_equal(x[0], modules[0].shape[1])
    assert_equal(x[1], modules[0].shape[0])


def test_segment_module_part():
    mod = data.datasets.poly10x6(1)[0]
    mod = detection.locate_module_and_cells(mod)
    part = detection.segment_module_part(mod, 1, 2, 2, 3)

    assert_equal(part.shape[1], 2 / 3 * part.shape[0], 0.1)
    assert part.first_col == 1
    assert part.first_row == 2
    assert part.cols == 2
    assert part.rows == 3


def test_segment_module_part_padding():
    mod = data.datasets.poly10x6(1)[0]
    mod = detection.locate_module_and_cells(mod)
    part = detection.segment_module_part(mod, -1, -2, 2, 3)

    assert_equal(part.shape[1], 2 / 3 * part.shape[0], 0.1)
    assert part.first_col == None
    assert part.first_row == None
    assert part.cols == 1
    assert part.rows == 1


def test_segment_size():
    img = detection.locate_module_and_cells(data.datasets.poly10x6(1)[0])
    part = detection.segment_module_part(img, 1, 3, 2, 1, size=20)
    assert_equal(part.shape[1], 2 * 20)
    assert_equal(part.shape[0], 1 * 20)


def test_segment_padding():
    img = detection.locate_module_and_cells(data.datasets.poly10x6(1)[0])
    part = detection.segment_module_part(img, 0, 0, 3, 2, padding=0.5)
    assert_equal(part.shape[1] / part.shape[0], 8 / 6)


def test_segment_padding_transform():
    img = detection.locate_module_and_cells(data.datasets.poly10x6(1)[0])
    part = detection.segment_module_part(img, 0, 0, 3, 2, padding=0.5, size=20)
    x1 = part.get_meta("transform")(np.array([[0.0, 1.0]])).flatten()
    x2 = part.get_meta("transform")(np.array([[2.0, 3.0]])).flatten()
    assert_equal(x1[0], 10)
    assert_equal(x1[1], 30)
    assert_equal(x2[0], 50)
    assert_equal(x2[1], 70)


def test_locate_partial_module():
    img = detection.locate_module_and_cells(data.datasets.poly10x6(1)[0])
    part = detection.segment_module_part(img, 0, 0, 3, 2, padding=0.5)
    part_det = detection.locate_module_and_cells(part, False)

    x1_true = [part_det.shape[1] * 1 / 8, part_det.shape[0] * 1 / 6]
    x2_true = [part_det.shape[1] * 7 / 8, part_det.shape[0] * 5 / 6]
    x1 = part_det.get_meta("transform")(np.array([[0.0, 0.0]])).flatten()
    x2 = part_det.get_meta("transform")(np.array([[3.0, 2.0]])).flatten()
    eps = 0.05 * part_det.shape[1]

    assert_equal(x1[0], x1_true[0], eps)
    assert_equal(x1[1], x1_true[1], eps)
    assert_equal(x2[0], x2_true[0], eps)
    assert_equal(x2[1], x2_true[1], eps)


def test_detect_multiple_modules():
    anns, imgs = data.datasets.multi_module_detection(N=2)
    modules, boxes = detection.locate_multiple_modules(
        imgs, return_bounding_boxes=True, padding=0.0
    )

    for path, prediction in boxes.items():
        _, _, recall = objdetect_metrics(anns[path.name], prediction, iou_thresh=[0.5])
        assert recall[0] > 0.5



class Summary:
    class __Summary:
        def __init__(self):
            self._values = {}
            self._enabled = False

        def enable(self):
            self._enabled = True

        def disable(self):
            self._enabled = False

        def put(self, k, v):
            if self._enabled:
                if k not in self._values.keys():
                    self._values[k] = v
                else:
                    raise RuntimeError("key {} already used".format(k))

        def get(self, k):
            return self._values.get(k)

        def flush(self):
            self._values = {}

        def keys(self):
            return self._values.keys()

    _instances = {}

    def __init__(self, key):
        if Summary._instances.get(key) is None:
            Summary._instances[key] = Summary.__Summary()
        self._key = key

    def __getattr__(self, name):
        return getattr(self._instances[self._key], name)



from pvinspect.data.datasets import (
    multi_module_detection,
    caip_dataB,
    caip_dataC,
    caip_dataD,
)
from pvinspect.common.util import objdetect_metrics
from pvinspect.preproc.detection import locate_multiple_modules, locate_module_and_cells
import numpy as np
from sklearn import metrics


def test_locate_multiple_modules_performance():
    # load data and perform detection with default parameters
    anns, imgs = multi_module_detection()
    modules, boxes = locate_multiple_modules(
        imgs, padding=0.0, return_bounding_boxes=True
    )

    # compute mean IOU over all images
    ious = list()
    for img in imgs:
        iou, _, _ = objdetect_metrics(anns[img.path.name], boxes[img.path])
        ious.append(iou)
    mean_iou = np.mean(ious)
    print(ious)

    # make sure that mean IOU > 0.8
    assert mean_iou > 0.8


CAIP_THRESH = [0.90, 0.91, 0.92, 0.93, 0.94, 0.95, 0.96, 0.97, 0.98, 0.99, 1.00]


def test_locate_module_and_cells_performance_caipB():
    # load data
    seq1, seq2, anns = caip_dataB()

    # perform detection
    seq1, boxes1 = locate_module_and_cells(
        seq1, estimate_distortion=False, return_bounding_boxes=True
    )
    seq2, boxes2 = locate_module_and_cells(
        seq2, estimate_distortion=False, return_bounding_boxes=True
    )

    # check IoU > 0.93
    ious = list()
    recalls = list()
    for img in seq1:
        iou, _, rec = objdetect_metrics(
            anns[img.path.name], boxes1[img.path.name], iou_thresh=CAIP_THRESH
        )
        ious.append(iou)
        recalls.append(rec)
    for img in seq2:
        iou, _, rec = objdetect_metrics(
            anns[img.path.name], boxes2[img.path.name], iou_thresh=CAIP_THRESH
        )
        ious.append(iou)
        recalls.append(rec)

    x = np.mean(recalls, axis=0)
    auc = metrics.auc(CAIP_THRESH, x)

    # TODO: Check, why this performs worse than in paper
    assert auc > 0.06


def test_locate_module_and_cells_performance_caipD():
    # load data
    seq, anns = caip_dataD()

    # perform detection
    seq, boxes = locate_module_and_cells(
        seq, estimate_distortion=False, return_bounding_boxes=True
    )

    # check IoU > 0.95
    ious = list()
    recalls = list()
    for img in seq:
        iou, _, rec = objdetect_metrics(
            anns[img.path.name], boxes[img.path.name], iou_thresh=CAIP_THRESH
        )
        ious.append(iou)
        recalls.append(rec)

    x = np.mean(recalls, axis=0)
    auc = metrics.auc(CAIP_THRESH, x)

    assert auc > 0.07



"""Detection, localization and segmentation of solar modules"""

import logging
from copy import deepcopy
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from pvinspect.common._ipy_exit import exit
from pvinspect.common.transform import (
    FullMultiTransform,
    FullTransform,
    HomographyTransform,
    warp_image,
)
from pvinspect.data.exceptions import UnsupportedModalityException
from pvinspect.data.image import *
from pvinspect.data.image import _sequence
from pvinspect.data.io import ObjectAnnotations
from pvinspect.preproc._mdetect.locate import apply
from shapely.geometry import Polygon
from skimage import filters, measure, morphology, transform
from skimage.filters.thresholding import threshold_otsu
from tqdm.auto import tqdm


@_sequence
def locate_module_and_cells(
    sequence: ModuleImageOrSequence,
    estimate_distortion: bool = True,
    orientation: str = None,
    return_bounding_boxes: bool = False,
) -> Union[Tuple[ModuleImageOrSequence, ObjectAnnotations], ModuleImageSequence]:
    """Locate a single module and its cells

    Note:
        This methods implements the following paper:
        Hoffmann, Mathis, et al. "Fast and robust detection of solar modules in electroluminescence images."
        International Conference on Computer Analysis of Images and Patterns. Springer, Cham, 2019.

    Args:
        sequence (ModuleImageOrSequence): A single module image or a sequence of module images 
        estimate_distortion (bool): Set True to estimate lens distortion, else False 
        orientation (str): Orientation of the module ('horizontal' or 'vertical' or None).
            If set to None (default), orientation is automatically determined
        return_bounding_boxes (bool): Indicates, if bounding boxes of returned modules are returned

    Returns:
        images: The same image/sequence with location information added
    """

    if sequence[0].modality != EL_IMAGE:
        logging.error("Module localization is not supporting given imaging modality")
        exit()

    result = list()
    failures = 0
    mcs = list()
    dts = list()
    flags = list()
    transforms = list()
    for img in tqdm(sequence.images):

        # very simple background suppression
        data = img.data.copy()
        thresh = threshold_otsu(data)
        data[data < thresh] = 0

        t, mc, dt, f = apply(
            data,
            img.cols,
            img.rows,
            is_module_detail=isinstance(img, PartialModuleImage),
            orientation=orientation,
        )
        transforms.append(t)
        flags.append(f)
        mcs.append(mc)
        dts.append(dt)

    if estimate_distortion:
        if sequence.same_camera:
            # do joint estimation
            logging.info(
                "Jointly estimating parameters for lens distortion. This might take some time.."
            )

            mcs_new = list()
            dts_new = list()
            valid = list()
            for mc, dt, f in zip(mcs, dts, flags):
                if mc is not None and dt is not None:
                    mcs_new.append(mc[f])
                    dts_new.append(dt[f])
                    valid.append(True)
                else:
                    valid.append(False)
            transforms = FullMultiTransform(
                mcs_new,
                dts_new,
                image_width=sequence.shape[1],
                image_height=sequence.shape[0],
                n_dist_coeff=1,
            )
            transforms_new = list()
            i = 0
            for v in valid:
                if v:
                    transforms_new.append(transforms[i])
                    i += 1
                else:
                    transforms_new.append(None)
            transforms = transforms_new

        else:
            transforms = list()
            for mc, dt, f, img in zip(mcs, dts, flags, sequence.images):
                if mc is not None and dt is not None:
                    t = FullTransform(
                        mc[f],
                        dt[f],
                        image_width=img.shape[1],
                        image_height=img.shape[0],
                        n_dist_coeff=1,
                    )
                    transforms.append(t)
                else:
                    transforms.append(None)

    for t, img in zip(transforms, sequence.images):
        if t is not None and t.valid:
            img_res = type(img).from_other(img, meta={"transform": t})
            result.append(img_res)
        else:
            result.append(deepcopy(img))
            failures += 1
    if failures > 0:
        logging.warning("Module localization falied for {:d} images".format(failures))

    result = ModuleImageSequence.from_other(sequence, images=result)

    if not return_bounding_boxes:
        return result
    else:
        boxes = dict()

        # compute polygon for every module and accumulate results
        for img in result:
            if img.has_meta("transform"):
                c = img.cols
                r = img.rows
                coords = np.array([[0.0, 0.0], [c, 0.0], [c, r], [0.0, r]])
                coords_transformed = img.get_meta("transform")(coords)
                poly = Polygon(
                    [
                        (x, y)
                        for x, y in zip(
                            coords_transformed[:, 0].tolist(),
                            coords_transformed[:, 1].tolist(),
                        )
                    ]
                )
                boxes[img.path.name] = [("Module", poly)]
            else:
                boxes[img.path.name] = []

        return result, boxes


def segment_module_part(
    image: Image,
    first_col: int,
    first_row: int,
    cols: int,
    rows: int,
    size: int = None,
    padding: float = 0.0,
) -> PartialModuleImage:
    """Segment a part of a module

    Args:
        image (Image): The corresponding image
        first_col (int): First column to appear in the segment
        first_row (int): First row to appear in the segment
        cols (int): Number of columns of the segment
        rows (int): Number of rows of the segment
        size (int): Size of a cell in pixels (automatically chosen by default)
        padding (float): Optional padding around the given segment relative to the cell size
                         (must be in [0..1[ )

    Returns:
        segment: The resulting segment
    """

    if not image.has_meta("transform") or not image.get_meta("transform").valid:
        logging.error(
            "The ModuleImage does not have a valid transform. Did module localization succeed?"
        )
        exit()

    t = image.get_meta("transform")

    if padding >= 1.0 or padding < 0.0:
        logging.error("padding needs to be in [0..1[")
        exit()

    last_col = first_col + cols
    last_row = first_row + rows

    size = t.mean_scale() if size is None else size
    result = warp_image(
        image.data,
        t,
        first_col - padding,
        first_row - padding,
        1 / size,
        1 / size,
        cols + 2 * padding,
        rows + 2 * padding,
    )
    result = result.astype(image.data.dtype)
    transform = HomographyTransform(
        np.array(
            [
                [first_col - padding, first_row - padding],
                [last_col + padding, first_row - padding],
                [last_col + padding, last_row + padding],
                [first_col - padding, last_row + padding],
            ]
        ),
        np.array(
            [
                [0.0, 0.0],
                [result.shape[1], 0.0],
                [result.shape[1], result.shape[0]],
                [0.0, result.shape[0]],
            ]
        ),
    )

    # bounding box in original image coords
    bb = [
        [first_col - padding, first_row - padding],
        [first_col + cols + padding, first_row + rows + padding],
    ]
    bb = t(np.array(bb))
    bb = Polygon.from_bounds(bb[0][0], bb[0][1], bb[1][0], bb[1][1])
    original = image.from_other(image, meta={"segment_module_original_box": bb})

    return PartialModuleImage.from_other(
        image,
        drop_meta_types=[Polygon],  # geometric attributes are invalid now..
        data=result,
        cols=cols + min(first_col, 0),
        rows=rows + min(first_row, 0),
        first_col=first_col if first_col >= 0 else None,
        first_row=first_row if first_row >= 0 else None,
        meta={"transform": transform, "segment_module_original": original},
    )


def segment_module(
    image: ModuleImage, size: int = None, padding: float = 0.0
) -> ModuleImage:
    """Obtain a rectified, cropped and undistorted module image

    Args:
        image (ModuleImage): A single module image
        size (int): Size of a cell in pixels (automatically chosen by default)
        padding (float): Optional padding around the given segment relative to the cell size
                         (must be in [0..1[ )

    Returns:
        module: The resulting module image
    """

    result = segment_module_part(image, 0, 0, image.cols, image.rows, size, padding)
    return ModuleImage.from_other(result)


def segment_cell(
    image: ModuleImage, row: int, col: int, size: int = None, padding: float = 0.0
) -> CellImage:
    """Obtain a cell image from a module image

    Args:
        image (ModuleImageOrSequence): A single module image
        row (int): The row number (starting at 0)
        col (int): The column number (starting at 0)
        size (int): Size of the resulting cell image in pixels (automatically chosen by default)
        padding (float): Optional padding around the cell relative to the cell size
                         (must be in [0..1[ )

    Returns:
        cells: The segmented cell image
    """

    result = segment_module_part(image, col, row, 1, 1, size, padding)
    return CellImage.from_other(result, row=row, col=col)


@_sequence
def segment_modules(
    sequence: ModuleImageOrSequence, size: int = None
) -> ModuleImageSequence:
    """Obtain rectified, cropped and undistorted module images from a sequence. Note that images that do not have a valid transform,
    possibly because the detection step failed, are silently ignored.

    Args:
        sequence (ModuleImageOrSequence): A single module image or a sequence of module images
        size (int): Size of the resulting cell images in pixels (automatically chosen by default)

    Returns:
        module: The segmented module images
    """

    scales = np.array(
        [
            img.get_meta("transform").mean_scale()
            for img in sequence.images
            if img.has_meta("transform") and img.get_meta("transform").valid
        ]
    )
    if scales.std() > 0.1 * scales.mean() and size is None:
        logging.warning(
            "The size of cells within the sequences varies by more than 10%. However, segment_modules, \
creates images of a fixed size. Please consider to split the sequence into multiple sequences \
with less variation in size."
        )
    if size is None:
        size = int(scales.mean())

    result = list()
    for img in tqdm(sequence.images):

        # for the moment, we silently ignore images without a valid transform
        if img.has_meta("transform") and img.get_meta("transform").valid:
            result.append(segment_module(img, size))

    return type(sequence).from_other(sequence, images=result, same_camera=False)


@_sequence(True)
def segment_cells(
    sequence: ModuleImageOrSequence, size: int = None
) -> CellImageSequence:
    """Obtain cell images from a sequence of module images. Note that images that do not have a valid transform,
    possibly because the detection step failed, are silently ignored.

    Args:
        sequence (ModuleImageOrSequence): A single module image or a sequence of module images
        size (int): Size of the resulting cell images in pixels (automatically chosen by default)

    Returns:
        cells: The segmented cell images
    """

    scales = np.array(
        [
            img.get_meta("transform").mean_scale()
            for img in sequence.images
            if img.has_meta("transform") and img.get_meta("transform").valid
        ]
    )
    if scales.std() > 0.1 * scales.mean() and size is None:
        logging.warning(
            "The size of cells within the sequences varies by more than 10%. However, segment_cells, \
creates cell images of a fixed size. Please consider to split the sequence into multiple sequences \
with less variation in size."
        )
    if size is None:
        size = int(scales.mean())

    result = list()
    for img in tqdm(sequence.images):
        for row in range(img.rows):
            for col in range(img.cols):

                # for the moment, we silently ignore images without a valid transform
                if (
                    img.has_meta("transform") is not None
                    and img.get_meta("transform").valid
                ):
                    result.append(segment_cell(img, row, col, size))

    return CellImageSequence(result)


def _do_locate_multiple_modules(
    image: Image,
    scale: float,
    reject_size_thresh: float,
    reject_fill_thresh: float,
    padding: float,
    cols: int,
    rows: int,
    drop_clipped_modules: bool,
) -> Tuple[List[ModuleImage], List[Polygon]]:

    # filter + binarize
    # image_f = filters.gaussian(image._data, filter_size)
    image_f = transform.rescale(image._data, scale)
    image_f = image_f > filters.threshold_otsu(image_f)

    # find regions
    labeled = morphology.label(image_f)
    regions = measure.regionprops(labeled)

    # process regions
    # check if bbox is filled to 100*reject_fill_thres%
    regions = [r for r in regions if r.area / r.bbox_area >= reject_fill_thresh]
    if len(regions) == 0:
        return [], []

    max_area = int(np.max([r.bbox_area for r in regions]))
    results = []
    boxes = []
    i = 0
    for r in regions:
        # check size
        if r.bbox_area < reject_size_thresh * max_area:
            continue

        # check not touching boundary
        if (
            r.bbox[0] == 0
            or r.bbox[1] == 0
            or r.bbox[2] == labeled.shape[0]
            or r.bbox[3] == labeled.shape[1]
        ) and drop_clipped_modules:
            continue

        # transform bounding box to original size
        s = 1 / scale
        bbox = [int(r.bbox[i] * s) for i in range(4)]

        # crop module
        pad = int(np.sqrt(r.bbox_area) * padding * s)
        y0, x0 = max(0, bbox[0] - pad), max(0, bbox[1] - pad)
        y1, x1 = (
            min(image.shape[0], bbox[2] + pad),
            min(image.shape[1], bbox[3] + pad),
        )
        boxes.append(("Module", Polygon.from_bounds(x0, y0, x1, y1)))
        crop = image._data[y0:y1, x0:x1]
        p = image.path.parent / "{}_module{:02d}{}".format(
            image.path.stem, i, image.path.suffix
        )
        results.append(
            ModuleImage(
                data=crop, modality=image.modality, path=p, cols=cols, rows=rows
            )
        )
        i += 1

    return results, boxes


@_sequence(True)
def locate_multiple_modules(
    sequence: ImageOrSequence,
    scale: float = 0.31,
    reject_size_thresh: float = 0.26,
    reject_fill_thresh: float = 0.42,
    padding: float = 0.05,
    drop_clipped_modules: bool = True,
    cols: int = None,
    rows: int = None,
    return_bounding_boxes: bool = False,
) -> Tuple[ModuleImageSequence, ObjectAnnotations]:
    """Perform localization and segmentation of multiple modules. The method is published in Hoffmann, Mathis, et al. 
    "Deep Learning-based Pipeline for Module Power Prediction from EL Measurements." arXiv preprint arXiv:2009.14712 (2020).

    Args:
        sequence (ImageOrSequence): Input images
        scale (float): Image is scaled to this size before processing
        reject_size_thresh (float): Detections smaller than this times the median size of detections are rejected
        reject_fill_thresh (float): Detections, where more that this parts of the area are black after thresholding are rejected
        padding (float): Detections are padded by this times the average size length of the bounding box
        drop_clipped_modules (bool): Indicate, if detections that touch the boundary are dropped
        cols (int): Number of columns of cells of a single module
        cols (rows): Number of rows of cells of a single module
        return_bounding_boxes (bool): If true, return the bounding boxes in addition to the crops

    Returns:
        The cropped modules as a ModuleImageSequence as well as (optionally), the bounding boxes.
    """

    # process sequence
    results = list()
    boxes = dict()
    # for img in tqdm(sequence):
    for img in tqdm(sequence):
        modules, b = _do_locate_multiple_modules(
            img,
            scale,
            reject_size_thresh,
            reject_fill_thresh,
            padding,
            cols,
            rows,
            drop_clipped_modules,
        )

        # add original images with box annotations as meta
        imgs_org = [
            Image.from_other(img, meta={"multimodule_index": i, "multimodule_boxes": b})
            for i in range(len(modules))
        ]
        modules = [
            ModuleImage.from_other(m, meta={"multimodule_original": o})
            for m, o in zip(modules, imgs_org)
        ]

        results += modules
        boxes[img.path] = b

    if return_bounding_boxes:
        if len(results) > 0:
            return ModuleImageSequence(results, same_camera=False), boxes
        else:
            return None, dict()
    else:
        if len(results) > 0:
            return ModuleImageSequence(results, same_camera=False)
        else:
            return None



import numpy as np
from pathlib import Path

from numpy import random
from pvinspect.data import (
    Image,
    ModuleImage,
    ImageSequence,
    ModuleImageSequence,
    EL_IMAGE,
)
from typing import List, Dict, Any, Optional
import uuid


def random_filename(ext: str = ".tif") -> str:
    return uuid.uuid4().hex + ext


def assert_equal(value, target, precision=1e-3):
    assert np.all(value > target - precision) and np.all(
        value < target + precision
    ), "got value={}, target={}".format(value, target)


def random_image(lazy: bool = False, **kwargs) -> Image:
    if lazy:
        data = Image.LazyData(lambda: np.random.random((10, 10)))
    else:
        data = np.random.random((10, 10))

    if "modality" not in kwargs.keys():
        kwargs["modality"] = EL_IMAGE
    if "path" not in kwargs.keys():
        kwargs["path"] = Path() / random_filename()

    return Image(data, **kwargs)


def random_uint_image(lazy: bool = False, **kwargs) -> Image:
    if lazy:
        data = Image.LazyData(
            lambda: (np.random.random((10, 10)) * 100).astype(np.uint32)
        )
    else:
        data = (np.random.random((10, 10)) * 100).astype(np.uint32)

    if "modality" not in kwargs.keys():
        kwargs["modality"] = EL_IMAGE
    if "path" not in kwargs.keys():
        kwargs["path"] = Path() / random_filename()

    return Image(data, **kwargs)


def random_ubyte_image(lazy: bool = False, **kwargs) -> Image:
    if lazy:
        data = Image.LazyData(
            lambda: (np.random.random((10, 10)) * 100).astype(np.uint8)
        )
    else:
        data = (np.random.random((10, 10)) * 100).astype(np.uint32)

    if "modality" not in kwargs.keys():
        kwargs["modality"] = EL_IMAGE
    if "path" not in kwargs.keys():
        kwargs["path"] = Path() / random_filename()

    return Image(data, **kwargs)


def random_module_image() -> ModuleImage:
    data = np.random.random((10, 10))
    return ModuleImage(data, modality=EL_IMAGE, path=Path() / random_filename())


def random_image_sequence(
    N: int = 3,
    meta: Optional[List[Dict[str, Any]]] = None,
    random_meta: bool = False,
    **kwargs
) -> ImageSequence:
    if random_meta and meta is None:
        random_keys = [uuid.uuid4().hex for _ in range(10)]
        meta = list()
        for _ in range(N):
            meta.append({k: random.randint(0, 10000) for k in random_keys})

    if meta is None:
        imgs = [random_image(**kwargs) for i in range(N)]
    else:
        assert len(meta) == N
        imgs = [random_image(meta=m, **kwargs) for m in meta]
    return ImageSequence(imgs, False)


def random_uint_image_sequence(
    N: int = 3, meta: Optional[List[Dict[str, Any]]] = None, **kwargs
) -> ImageSequence:
    if meta is None:
        imgs = [random_uint_image(**kwargs) for i in range(N)]
    else:
        assert len(meta) == N
        imgs = [random_uint_image(meta=m, **kwargs) for m in meta]
    return ImageSequence(imgs, False)


def random_ubyte_image_sequence(
    N: int = 3, meta: Optional[List[Dict[str, Any]]] = None, **kwargs
) -> ImageSequence:
    if meta is None:
        imgs = [random_ubyte_image(**kwargs) for i in range(N)]
    else:
        assert len(meta) == N
        imgs = [random_ubyte_image(meta=m, **kwargs) for m in meta]
    return ImageSequence(imgs, False)


def random_module_image_sequence() -> ModuleImageSequence:
    imgs = [random_module_image() for x in range(3)]
    return ModuleImageSequence(imgs, False)



from pvinspect.data.io import *
from pvinspect.data.io import (
    _prepare_json_meta,
    _load_json_meta_hook,
    _get_meta_cache_path,
)
from pvinspect.data import datasets
import pvinspect.data as data
from pvinspect.preproc.detection import locate_module_and_cells, segment_cells
from pathlib import Path
import numpy as np
from pvinspect.data.image import *
from skimage.io import imsave, imread
from test.utilities import *
import json
import datetime
from shapely.geometry import Polygon, Point

EXAMPLES = (
    Path(__file__).absolute().parent.parent.parent
    / "pvinspect"
    / "data"
    / "datasets"
    / "20191219_poly10x6"
)


def _check_download_demo():
    data.datasets.poly10x6(1)


def _test_dict():
    return {
        "a": 1,
        "b": {"a": "a", "b": datetime.datetime.now()},
        "c": Polygon.from_bounds(0, 0, 10, 5),
        "d": Point(10, 20),
        "e": np.zeros([10]),
        "f": Modality.EL_IMAGE,
    }


def test_read_sequence():
    _check_download_demo()
    seq = read_module_images(EXAMPLES, EL_IMAGE, True)
    assert len(seq) == 20
    assert seq.modality == EL_IMAGE


def test_limit():
    _check_download_demo()
    seq = read_module_images(EXAMPLES, EL_IMAGE, False, N=2)
    assert len(seq) == 2


def test_same_camera():
    _check_download_demo()
    seq = read_module_images(EXAMPLES, EL_IMAGE, True, N=1)
    assert seq.same_camera == True


def test_different_dtypes():
    _check_download_demo()
    seq = read_module_images(EXAMPLES, EL_IMAGE, False, allow_different_dtypes=True)
    assert seq.dtype == None


def test_dtype():
    _check_download_demo()
    seq = read_module_images(EXAMPLES, EL_IMAGE, False)
    assert seq.dtype == DType.UNSIGNED_INT


def test_shape():
    _check_download_demo()
    seq = read_module_images(EXAMPLES, EL_IMAGE, True)
    assert seq.shape == (2052, 2046)


def test_shape_not_same_camera():
    _check_download_demo()
    seq = read_module_images(EXAMPLES, EL_IMAGE, False)
    assert seq.shape == None


def test_filter():
    _check_download_demo()
    seq = read_module_images(EXAMPLES, EL_IMAGE, True, pattern="*.tif", N=1)
    assert len(seq) == 1
    seq = read_module_images(EXAMPLES, EL_IMAGE, True, pattern=("*.png", "*.tif"), N=1)
    assert len(seq) == 1
    try:
        seq = read_module_images(EXAMPLES, EL_IMAGE, True, pattern="*.png", N=1)
        assert False
    except:
        assert True


def test_save_and_read_image(tmp_path):
    img = datasets.poly10x6(1)[0]
    save_image(tmp_path / "img.tif", img)
    img_read = read_module_image(tmp_path / "img.tif", EL_IMAGE)
    assert np.linalg.norm(img.data.flatten() - img_read.data.flatten()) == 0


def test_save_image_sequence(tmp_path):
    seq = datasets.poly10x6(5)
    save_images(tmp_path, seq)

    for img in seq:
        img_read = read_module_image(tmp_path / img.path.name, EL_IMAGE)


def test_save_cell_images(tmp_path):
    cells = segment_cells(locate_module_and_cells(datasets.poly10x6(1)[0]))
    save_images(tmp_path, cells)

    for cell in cells:
        p = tmp_path / "{}_row{:02d}_col{:02d}{}".format(
            cell.path.stem, cell.row, cell.col, cell.path.suffix
        )
        img_read = read_module_image(p, EL_IMAGE)


def test_read_images():
    _check_download_demo()
    seq = read_images(EXAMPLES, True, N=2)
    assert isinstance(seq, ImageSequence)
    for img in seq:
        assert isinstance(img, Image)


def test_lazy_read_images():
    _check_download_demo()
    seq = read_images(EXAMPLES, True, N=2, lazy=True)
    assert isinstance(seq[0]._data, Image.LazyData)


def test_read_module_images():
    _check_download_demo()
    seq = read_module_images(EXAMPLES, EL_IMAGE, True, N=2)
    assert isinstance(seq, ModuleImageSequence)
    for img in seq:
        assert isinstance(img, ModuleImage)


def test_read_partial_module_images():
    _check_download_demo()
    seq = read_partial_module_images(EXAMPLES, EL_IMAGE, True, N=2)
    assert isinstance(seq, ModuleImageSequence)
    for img in seq:
        assert isinstance(img, PartialModuleImage)


def test_save_image_with_visualization(tmp_path: Path):
    img = datasets.poly10x6(1)[0]
    p = tmp_path / "img.pdf"
    save_image(p, img, with_visusalization=True)
    assert p.is_file()


def test_save_images_filename_hook(tmp_path: Path):
    cells = segment_cells(locate_module_and_cells(datasets.poly10x6(1)[0]))
    hook = lambda x: "{}_{:02d}{:02d}{}".format(
        x.path.stem, x.row, x.col, x.path.suffix
    )
    save_images(tmp_path, cells)

    for cell in cells:
        p = tmp_path / "{}_{:02d}{:02d}{}".format(
            cell.path.stem, cell.row, cell.col, cell.path.suffix
        )
        # try read
        img_read = read_module_image(p, EL_IMAGE)


def test_force_dtype(tmp_path: Path):
    # create samples
    a = np.random.rand(10, 10) * 100
    imsave(tmp_path / "wrong_float.tif", a)

    seq = read_image(tmp_path / "wrong_float.tif", force_dtype=DType.UNSIGNED_INT)
    assert seq.dtype == DType.UNSIGNED_INT


def test_save_float_image_conversion(tmp_path: Path):
    img = random_image()
    img = Image.from_other(img, data=img.data.astype(np.float64))
    save_image(tmp_path / "test.tif", img)

    img = imread(tmp_path / "test.tif")
    assert img.dtype == np.float32


def test_hierachical_with_keys_save(tmp_path: Path):
    img1 = random_image(meta={"m1": 1, "m2": 1})
    img2 = random_image(meta={"m1": 1, "m2": 2})
    img3 = random_image(meta={"m1": 2, "m2": 1})
    seq = ImageSequence([img1, img2, img3], same_camera=True)

    save_images(tmp_path, seq, hierarchical=["m1", "m2"])

    for img in seq:
        p = (
            tmp_path
            / ("m1_" + str(img.get_meta("m1")))
            / ("m2_" + str(img.get_meta("m2")))
            / img.path.name
        )
        ref = read_image(p, lazy=False)
        assert_equal(img.data, ref.data)


def test_hierachical_without_keys_save(tmp_path: Path):
    img1 = random_image(meta={"m1": 1, "m2": 1})
    img2 = random_image(meta={"m1": 1, "m2": 2})
    img3 = random_image(meta={"m1": 2, "m2": 1})
    seq = ImageSequence([img1, img2, img3], same_camera=True)

    save_images(tmp_path, seq, hierarchical=["m1", "m2"], include_meta_keys=False)

    for img in seq:
        p = tmp_path / str(img.get_meta("m1")) / str(img.get_meta("m2")) / img.path.name
        ref = read_image(p, lazy=False)
        assert_equal(img.data, ref.data)


def test_prepare_json():
    data = _test_dict()
    data = _prepare_json_meta(data)

    assert data["e"] is None
    assert isinstance(data["d"], str)
    assert isinstance(data["c"], str)
    assert isinstance(data["b"], dict)
    assert isinstance(data["a"], int)
    assert isinstance(data["b"]["b"], str)


def test_dump_load_json():
    data = _test_dict()
    s = json.dumps(_prepare_json_meta(data))
    result = json.loads(s, object_hook=_load_json_meta_hook)

    data["e"] = None
    assert result == data


def test_save_and_load_image_with_meta(tmp_path: Path):
    meta = {"a": 1, "b": "xxx"}
    img = random_image(meta=meta)
    save_image(tmp_path / "test.tif", img, save_meta=True)
    img2 = read_image(tmp_path / "test.tif")

    d1 = img.meta_to_pandas().to_dict()
    d2 = img2.meta_to_pandas().to_dict()
    del d1["path"]
    del d2["path"]
    assert d1 == d2


def test_save_with_prefix(tmp_path: Path):
    imgs = random_image_sequence()
    save_images(tmp_path, imgs, filename_prefix="XY_")

    for img in imgs:
        fn = "XY_" + img.path.name
        assert (tmp_path / fn).is_file()


def test_save_with_suffix(tmp_path: Path):
    imgs = random_image_sequence()
    save_images(tmp_path, imgs, filename_suffix="_XY")

    for img in imgs:
        fn = img.path.stem + "_XY" + img.path.suffix
        assert (tmp_path / fn).is_file()


def test_meta_cache_is_created(tmp_path: Path):
    images = random_image_sequence(random_meta=True)
    save_images(tmp_path, images, save_meta=True)
    images_loaded = read_images(tmp_path)

    # check if cache is created
    assert _get_meta_cache_path(tmp_path).is_file()

    # loading again should reveal the same meta data
    assert images_loaded.meta_to_pandas().equals(read_images(tmp_path).meta_to_pandas())


def test_meta_cache_is_invalidated(tmp_path: Path):
    images = random_image_sequence(random_meta=True)
    save_images(tmp_path, images, save_meta=True)
    images_loaded = read_images(tmp_path)

    # modify one key of a single image
    target_image = images_loaded[0]
    target_key = target_image.meta_to_pandas().keys()[0]
    target_image = target_image.from_self(meta={target_key: 1})
    save_image(target_image.path, target_image, save_meta=True)

    # load and test
    target_image_loaded = read_images(tmp_path)[0]
    assert target_image_loaded.meta_to_pandas().equals(target_image.meta_to_pandas())



import numpy as np
from shapely.geometry import Polygon
from typing import List
from contextlib import contextmanager
import sys, os


def rotation_matrix_3d(rx, ry, rz):
    Rx = np.array(
        [[1, 0, 0], [0, np.cos(rx), -np.sin(rx)], [0, np.sin(rx), np.cos(rx)]]
    )
    Ry = np.array(
        [[np.cos(ry), 0, np.sin(ry)], [0, 1, 0], [-np.sin(ry), 0, np.cos(ry)]]
    )
    Rz = np.array(
        [[np.cos(rz), -np.sin(rz), 0], [np.sin(rz), np.cos(rz), 0], [0, 0, 1]]
    )
    return Rz.dot(Ry.dot(Rx))


def random_rotation_matrix_3d(max_rotation):
    rot = (np.random.rand(3) * 2 * max_rotation) - max_rotation
    return rotation_matrix_3d(rot[0], rot[1], rot[2])


def random_rotation_translation_matrix_3d(
    ranges=[(-20, 20), (-20, 20), (200, 400)], max_rotation=np.pi
):
    ranges = np.array(ranges)
    scales = ranges[:, 1] - ranges[:, 0]
    t = (np.random.rand(3) * scales + ranges[:, 0]).reshape(3, 1)
    r = random_rotation_matrix_3d(max_rotation)
    return np.concatenate([r, t], axis=1)


def rodrigues2matrix(r):
    angle = np.linalg.norm(r)
    r = r / angle
    W = np.array(
        [[0, -r[2], r[1]], [r[2], 0, -r[0]], [-r[1], r[0], 0]], dtype=np.float64
    )
    r = r.reshape(-1, 1)
    rrT = r.dot(r.T)
    return (
        np.cos(angle) * np.identity(3) + (1 - np.cos(angle)) * rrT + np.sin(angle) * W
    )


def matrix2rodrigues(R):
    A = (R - R.T) / 2
    p = np.array([A[2, 1], A[0, 2], A[1, 0]])
    pn = np.linalg.norm(p)
    c = (np.trace(R) - 1) / 2

    if pn == 0.0 and c == 1.0:
        return np.zeros(3)
    elif pn == 0.0 and c == -1.0:
        Rp = R + np.identity(3)
        v = R[:, np.argmax(np.linalg.norm(R, axis=1))]
        u = v / np.linalg.norm(v)
        Su = (
            -u
            if u[0] < 0.0
            or (u[0] == 0.0 and u[1] < 0.0)
            or (u[0] == 0.0 and u[1] == 0.0 and u[2] < 0.0)
            else u
        )
        return np.pi * Su
    else:
        u = p / pn
        angle = np.arctan2(pn, c)
        return angle * u


def ransac(n_samples, n_required, model_f, err_f, p=0.99):
    is_inlier = np.zeros(n_samples, dtype=np.bool)
    num_inlier = 0
    N = np.inf
    n = 0
    best_model = None

    while n < N:
        # randomly sample n_required points
        idx = np.random.choice(n_samples, n_required, replace=False)
        mask = np.full_like(is_inlier, False)
        mask[idx] = True

        # determine model
        # returns None if failed
        tmp = model_f(mask)

        if tmp is not None:
            # compute errors
            new_inlier = err_f(tmp)
            num_new_inlier = np.sum(new_inlier)

            # accept result?
            if num_new_inlier > num_inlier:
                best_model = tmp
                is_inlier = new_inlier
                num_inlier = num_new_inlier
                w = num_inlier / n_samples
                N = np.log(1 - p) / np.log(1 - w ** n_required + 1e-8)

        n += 1

    # compute model using all inliers
    return model_f(is_inlier), is_inlier, n


def filter_lists(lists, indicator):
    res = [[v for i, v in enumerate(l) if indicator[i]] for l in lists]
    return tuple(res)


def find_file_in_list(file, files):
    idx = None
    for i, p in enumerate(files):
        if file.samefile(p):
            return i

    return None


def iou(p1, p2):
    """Calculate IOU of two shapely polygons"""
    if p1.intersects(p2):
        return p1.intersection(p2).area / p1.union(p2).area
    else:
        return 0.0


def objdetect_metrics(
    objects_label,
    objects_predicted,
    iou_thresh: List[float] = [0.6, 0.7, 0.8, 0.9, 0.99],
):
    """Calculate IOU for every pair from objects_label and objects_predicted,
    assign objects by maximum IOU and return mean IOU, precision and recall"""

    t = objects_label[0][0]
    for o in objects_label + objects_predicted:
        if o[0] != t:
            raise NotImplementedError("Method does not support different object types")

    # get rid of object type
    objects_label = [v[1] for v in objects_label]
    objects_predicted = [v[1] for v in objects_predicted]

    # set up matrix of IOUs
    max_len = max(len(objects_label), len(objects_predicted))
    ious = np.zeros((max_len, max_len), dtype=np.float64)
    for i, o1 in enumerate(objects_label):
        for j, o2 in enumerate(objects_predicted):
            ious[i, j] = iou(o1, o2)

    # for every object from objects1, find partner with maximum intersection from object2
    iou_sum = 0.0
    assignments = [dict() for _ in iou_thresh]
    for i in range(max_len):
        # find detection with maximum IOU to label i
        j = np.argmax(ious[i])

        # only assign, if j does not have a larger IOU with another label
        if np.argmax(ious[:, j]) == i:
            iou_sum += ious[i, j]

            # store assignments by iou_thresh
            for thresh_i, thresh in enumerate(iou_thresh):
                if ious[i, j] > thresh:
                    # i = label index; j = detection index
                    assignments[thresh_i][i] = j

            # "delete" partner object
            ious[:, j] = 0.0

    # calculate precision + recall
    recall = [
        len(x.keys()) / len(objects_label) if len(objects_label) > 0 else 0.0
        for x in assignments
    ]
    precision = [
        len(x.keys()) / len(objects_predicted) if len(objects_predicted) > 0 else 0.0
        for x in assignments
    ]
    return iou_sum / max_len, precision, recall


def polygon2boundingbox(polygon: Polygon):
    """Convert polygon to bounding box"""
    return Polygon.from_bounds(*polygon.bounds)


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


@contextmanager
def suppress_stderr():
    with open(os.devnull, "w") as devnull:
        old_stderr = sys.stderr
        sys.stderr = devnull
        try:
            yield
        finally:
            sys.stderr = old_stderr



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from scipy.ndimage.filters import gaussian_filter1d
import math
from .config import (
    GAUSSIAN_RELATIVE_SIGMA,
    OUTER_CORNER_THRESH_FACTOR,
    MODULE_DETECTION_PEAK_THRESH,
)
from scipy import optimize, signal
from pvinspect.common import transform
from .summary import Summary

summary = Summary("locate_module")


def _find_stops(img, dim, n_cells):
    img = img.T if dim == 0 else img

    # calculate downsampling
    size = 0.5 * np.sum(img.shape)

    # extract profile of cumsum along di
    profile = np.sum(img, 1)
    profile_smooth = gaussian_filter1d(profile, GAUSSIAN_RELATIVE_SIGMA * size)
    profile_smooth = profile_smooth - np.min(profile_smooth)
    profile_smooth = profile_smooth / (np.max(profile_smooth) + 1e-5)

    # calculate gradient of that
    grad_smooth = np.gradient(profile_smooth)

    thresh = MODULE_DETECTION_PEAK_THRESH * np.mean(np.abs(grad_smooth))
    peaks_max, _ = signal.find_peaks(grad_smooth, height=thresh)
    peaks_min, _ = signal.find_peaks(-grad_smooth, height=thresh)

    if len(peaks_max) == 0 or len(peaks_min) == 0:
        return None

    extremals = [peaks_max[0], peaks_min[-1]]

    thresh = np.std(grad_smooth) * OUTER_CORNER_THRESH_FACTOR

    min_distance = int(
        (extremals[1] - extremals[0]) / n_cells / 2
    )  # consider std only for half cell size

    res = []
    thresh = (
        np.std(
            np.clip(grad_smooth, 0.0, None)[
                max(0, extremals[0] - min_distance) : min(
                    img.shape[0], extremals[0] + min_distance
                )
            ]
        )
        * OUTER_CORNER_THRESH_FACTOR
    )
    res.append(extremals[0] - np.argmax((grad_smooth <= thresh)[extremals[0] :: -1]))
    res.append(extremals[0] + np.argmax((grad_smooth <= thresh)[extremals[0] :: +1]))
    thresh = (
        np.std(
            np.clip(grad_smooth, None, 0.0)[
                max(0, extremals[1] - min_distance) : min(
                    img.shape[0], extremals[1] + min_distance
                )
            ]
        )
        * OUTER_CORNER_THRESH_FACTOR
    )
    res.append(extremals[1] - np.argmax((grad_smooth >= -thresh)[extremals[1] :: -1]))
    res.append(extremals[1] + np.argmax((grad_smooth >= -thresh)[extremals[1] :: +1]))

    # plt.plot(np.arange(profile.shape[0]), grad_smooth)
    # plt.scatter(peaks_max, grad_smooth[peaks_max])
    # plt.scatter(peaks_min, grad_smooth[peaks_min])
    # plt.scatter(res, [grad_smooth[x] for x in res])
    # plt.show()

    return res


def _assign_stops(x_stops, y_stops, img):

    # find outer and inner bounding box
    outer_anchor = (min(x_stops[0], x_stops[1]), min(y_stops[0], y_stops[1]))
    inner_anchor = (max(x_stops[0], x_stops[1]), max(y_stops[0], y_stops[1]))
    outer_size = (
        max(x_stops[2], x_stops[3]) - outer_anchor[0] + 1,
        max(y_stops[2], y_stops[3]) - outer_anchor[1] + 1,
    )
    inner_size = (
        min(x_stops[2], x_stops[3]) - inner_anchor[0] + 1,
        min(y_stops[2], y_stops[3]) - inner_anchor[1] + 1,
    )
    # fig, ax = plt.subplots(1)
    # ax.imshow(img, cmap='gray')
    # rect_outer = patches.Rectangle(outer_anchor, *outer_size, linewidth=1, edgecolor='r', facecolor='none')
    # rect_inner = patches.Rectangle(inner_anchor, *inner_size, linewidth=1, edgecolor='r', facecolor='none')
    # ax.add_patch(rect_outer)
    # ax.add_patch(rect_inner)
    # plt.show()

    # find boxes between outer and inner
    upper_margin = (
        max(x_stops[0], x_stops[1]),
        min(y_stops[0], y_stops[1]),
        min(x_stops[2], x_stops[3]),
        max(y_stops[0], y_stops[1]),
    )
    left_margin = (
        min(x_stops[0], x_stops[1]),
        max(y_stops[0], y_stops[1]),
        max(x_stops[0], x_stops[1]),
        min(y_stops[2], y_stops[3]),
    )
    bottom_margin = (
        max(x_stops[0], x_stops[1]),
        min(y_stops[2], y_stops[3]),
        min(x_stops[2], x_stops[3]),
        max(y_stops[2], y_stops[3]),
    )
    right_margin = (
        min(x_stops[2], x_stops[3]),
        max(y_stops[0], y_stops[1]),
        max(x_stops[2], x_stops[3]),
        min(y_stops[2], y_stops[3]),
    )

    # divide them into left/right or top/bottom part
    upper_margin_l = (
        upper_margin[1],
        upper_margin[0],
        upper_margin[3],
        upper_margin[0] + (upper_margin[2] - upper_margin[0] + 1) // 2,
    )
    upper_margin_r = (
        upper_margin[1],
        upper_margin_l[3],
        upper_margin[3],
        upper_margin[2],
    )
    left_margin_t = (
        left_margin[1],
        left_margin[0],
        left_margin[1] + (left_margin[3] - left_margin[1] + 1) // 2,
        left_margin[2],
    )
    left_margin_b = (left_margin_t[2], left_margin_t[1], left_margin[3], left_margin[2])
    bottom_margin_l = (
        bottom_margin[1],
        bottom_margin[0],
        bottom_margin[3],
        bottom_margin[0] + (bottom_margin[2] - bottom_margin[0] + 1) // 2,
    )
    bottom_margin_r = (
        bottom_margin[1],
        bottom_margin_l[3],
        bottom_margin[3],
        bottom_margin[2],
    )
    right_margin_t = (
        right_margin[1],
        right_margin[0],
        right_margin[1] + (right_margin[3] - right_margin[1] + 1) // 2,
        right_margin[2],
    )
    right_margin_b = (
        right_margin_t[2],
        right_margin[0],
        right_margin[3],
        right_margin[2],
    )

    summary.put("box_upper_l", upper_margin_l)
    summary.put("box_upper_r", upper_margin_r)
    summary.put("box_left_t", left_margin_t)
    summary.put("box_left_b", left_margin_b)
    summary.put("box_bottom_l", bottom_margin_l)
    summary.put("box_bottom_r", bottom_margin_r)
    summary.put("box_right_t", right_margin_t)
    summary.put("box_right_b", right_margin_b)

    def sum_helper(img, p0, p1):
        s0 = int(np.ceil((p1[0] - p0[0]) / 10))  # sample approx. 10 lines per dim
        s1 = int(np.ceil((p1[1] - p0[1]) / 10))
        s0 = s0 if s0 > 0 else 1
        s1 = s1 if s1 > 0 else 1
        return np.sum(img[p0[0] : p1[0] + 1 : s0, p0[1] : p1[1] + 1 : s1])

    # sum over parts
    sum_upper_margin_l = sum_helper(img, upper_margin_l[:2], upper_margin_l[2:])
    sum_upper_margin_r = sum_helper(img, upper_margin_r[:2], upper_margin_r[2:])
    sum_left_margin_t = sum_helper(img, left_margin_t[:2], left_margin_t[2:])
    sum_left_margin_b = sum_helper(img, left_margin_b[:2], left_margin_b[2:])
    sum_bottom_margin_l = sum_helper(img, bottom_margin_l[:2], bottom_margin_l[2:])
    sum_bottom_margin_r = sum_helper(img, bottom_margin_r[:2], bottom_margin_r[2:])
    sum_right_margin_t = sum_helper(img, right_margin_t[:2], right_margin_t[2:])
    sum_right_margin_b = sum_helper(img, right_margin_b[:2], right_margin_b[2:])

    # assign stops
    if sum_upper_margin_l > sum_upper_margin_r:
        Ay = y_stops[0]
        By = y_stops[1]
    else:
        Ay = y_stops[1]
        By = y_stops[0]

    if sum_bottom_margin_l > sum_bottom_margin_r:
        Cy = y_stops[2]
        Dy = y_stops[3]
    else:
        Cy = y_stops[3]
        Dy = y_stops[2]

    if sum_left_margin_t > sum_left_margin_b:
        Ax = x_stops[0]
        Dx = x_stops[1]
    else:
        Ax = x_stops[1]
        Dx = x_stops[0]

    if sum_right_margin_t > sum_right_margin_b:
        Bx = x_stops[3]
        Cx = x_stops[2]
    else:
        Bx = x_stops[2]
        Cx = x_stops[3]

    return np.array([(Ax, Ay), (Bx, By), (Cx, Cy), (Dx, Dy)])


def locate_module(img, n_cols, m_rows):
    x_stops = _find_stops(img, 0, n_cols)
    y_stops = _find_stops(img, 1, m_rows)

    if x_stops is None or y_stops is None:
        return None
    else:
        return _assign_stops(x_stops, y_stops, img)


def module_boundingbox_model(coords, n_cols, m_rows, orientation):
    if orientation is None:
        mean_x = 1 / 2 * (coords[1, 0] - coords[0, 0] + coords[2, 0] - coords[3, 0])
        mean_y = 1 / 2 * (coords[2, 1] - coords[1, 1] + coords[3, 1] - coords[0, 1])
        oriented_horizontal = mean_x > mean_y
    else:
        oriented_horizontal = orientation == "horizontal"

    if oriented_horizontal:
        src = np.array([[0, 0], [n_cols, 0], [n_cols, m_rows], [0, m_rows]])
    else:
        src = np.array([[0, m_rows], [0, 0], [n_cols, 0], [n_cols, m_rows]])

    return src
